# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("IncompatibleGraphError", "QuantumGraph")

import datetime
import getpass
import io
import json
import lzma
import os
import struct
import sys
import time
import uuid
from collections import defaultdict, deque
from collections.abc import Generator, Iterable, Iterator, Mapping, MutableMapping
from itertools import chain
from types import MappingProxyType
from typing import Any, BinaryIO, TypeVar

import networkx as nx
from networkx.drawing.nx_agraph import write_dot

import lsst.utils.logging
from lsst.daf.butler import (
    Config,
    DatasetId,
    DatasetRef,
    DatasetType,
    DimensionRecordsAccumulator,
    DimensionUniverse,
    LimitedButler,
    Quantum,
    QuantumBackedButler,
)
from lsst.daf.butler._rubin import generate_uuidv7
from lsst.daf.butler.datastore.record_data import DatastoreRecordData
from lsst.daf.butler.persistence_context import PersistenceContextVars
from lsst.daf.butler.registry import ConflictingDefinitionError
from lsst.resources import ResourcePath, ResourcePathExpression
from lsst.utils.introspection import get_full_type_name
from lsst.utils.packages import Packages

from ..config import PipelineTaskConfig
from ..connections import iterConnections
from ..pipeline import TaskDef
from ..pipeline_graph import PipelineGraph, compare_packages, log_config_mismatch
from ._implDetails import DatasetTypeName, _DatasetTracker
from ._loadHelpers import LoadHelper
from ._versionDeserializers import DESERIALIZER_MAP
from .graphSummary import QgraphSummary, QgraphTaskSummary
from .quantumNode import BuildId, QuantumNode

_T = TypeVar("_T", bound="QuantumGraph")
_LOG = lsst.utils.logging.getLogger(__name__)

# modify this constant any time the on disk representation of the save file
# changes, and update the load helpers to behave properly for each version.
SAVE_VERSION = 3

# Strings used to describe the format for the preamble bytes in a file save
# The base is a big endian encoded unsigned short that is used to hold the
# file format version. This allows reading version bytes and determine which
# loading code should be used for the rest of the file
STRUCT_FMT_BASE = ">H"
#
# Version 1
# This marks a big endian encoded format with an unsigned short, an unsigned
# long long, and an unsigned long long in the byte stream
# Version 2
# A big endian encoded format with an unsigned long long byte stream used to
# indicate the total length of the entire header.
STRUCT_FMT_STRING = {1: ">QQ", 2: ">Q"}

# magic bytes that help determine this is a graph save
MAGIC_BYTES = b"qgraph4\xf6\xe8\xa9"


class IncompatibleGraphError(Exception):
    """Exception class to indicate that a lookup by NodeId is impossible due
    to incompatibilities.
    """

    pass


class QuantumGraph:
    """QuantumGraph is a directed acyclic graph of `QuantumNode` objects.

    This data structure represents a concrete workflow generated from a
    `Pipeline`.

    Parameters
    ----------
    quanta : `~collections.abc.Mapping` [ `TaskDef`, \
            `set` [ `~lsst.daf.butler.Quantum` ] ]
        This maps tasks (and their configs) to the sets of data they are to
        process.
    metadata : Optional `~collections.abc.Mapping` of `str` to primitives
        This is an optional parameter of extra data to carry with the graph.
        Entries in this mapping should be able to be serialized in JSON.
    universe : `~lsst.daf.butler.DimensionUniverse`, optional
        The dimensions in which quanta can be defined. Need only be provided if
        no quanta have data IDs.
    initInputs : `~collections.abc.Mapping`, optional
        Maps tasks to their InitInput dataset refs. Dataset refs can be either
        resolved or non-resolved. Presently the same dataset refs are included
        in each `~lsst.daf.butler.Quantum` for the same task.
    initOutputs : `~collections.abc.Mapping`, optional
        Maps tasks to their InitOutput dataset refs. Dataset refs can be either
        resolved or non-resolved. For intermediate resolved refs their dataset
        ID must match ``initInputs`` and Quantum ``initInputs``.
    globalInitOutputs : `~collections.abc.Iterable` \
            [ `~lsst.daf.butler.DatasetRef` ], optional
        Dataset refs for some global objects produced by pipeline. These
        objects include task configurations and package versions. Typically
        they have an empty DataId, but there is no real restriction on what
        can appear here.
    registryDatasetTypes : `~collections.abc.Iterable` \
            [ `~lsst.daf.butler.DatasetType` ],  optional
        Dataset types which are used by this graph, their definitions must
        match registry. If registry does not define dataset type yet, then
        it should match one that will be created later.

    Raises
    ------
    ValueError
        Raised if the graph is pruned such that some tasks no longer have nodes
        associated with them.
    """

    def __init__(
        self,
        quanta: Mapping[TaskDef, set[Quantum]],
        metadata: Mapping[str, Any] | None = None,
        universe: DimensionUniverse | None = None,
        initInputs: Mapping[TaskDef, Iterable[DatasetRef]] | None = None,
        initOutputs: Mapping[TaskDef, Iterable[DatasetRef]] | None = None,
        globalInitOutputs: Iterable[DatasetRef] | None = None,
        registryDatasetTypes: Iterable[DatasetType] | None = None,
    ):
        self._buildGraphs(
            quanta,
            metadata=metadata,
            universe=universe,
            initInputs=initInputs,
            initOutputs=initOutputs,
            globalInitOutputs=globalInitOutputs,
            registryDatasetTypes=registryDatasetTypes,
        )

    def _buildGraphs(
        self,
        quanta: Mapping[TaskDef, set[Quantum]],
        *,
        _quantumToNodeId: Mapping[Quantum, uuid.UUID] | None = None,
        _buildId: BuildId | None = None,
        metadata: Mapping[str, Any] | None = None,
        universe: DimensionUniverse | None = None,
        initInputs: Mapping[TaskDef, Iterable[DatasetRef]] | None = None,
        initOutputs: Mapping[TaskDef, Iterable[DatasetRef]] | None = None,
        globalInitOutputs: Iterable[DatasetRef] | None = None,
        registryDatasetTypes: Iterable[DatasetType] | None = None,
    ) -> None:
        """Build the graph that is used to store the relation between tasks,
        and the graph that holds the relations between quanta
        """
        # Save packages to metadata
        self._metadata = dict(metadata) if metadata is not None else {}
        self._metadata.setdefault("packages", Packages.fromSystem())
        self._metadata.setdefault("user", getpass.getuser())
        self._metadata.setdefault("time", f"{datetime.datetime.now()}")
        self._metadata.setdefault("full_command", " ".join(sys.argv))

        self._buildId = _buildId if _buildId is not None else BuildId(f"{time.time()}-{os.getpid()}")
        # Data structure used to identify relations between
        # DatasetTypeName -> TaskDef.
        self._datasetDict = _DatasetTracker(createInverse=True)

        # Temporary graph that will have dataset UUIDs (as raw bytes) and
        # QuantumNode objects as nodes; will be collapsed down to just quanta
        # later.
        bipartite_graph = nx.DiGraph()

        self._nodeIdMap: dict[uuid.UUID, QuantumNode] = {}
        self._taskToQuantumNode: MutableMapping[TaskDef, set[QuantumNode]] = defaultdict(set)
        for taskDef, quantumSet in quanta.items():
            connections = taskDef.connections

            # For each type of connection in the task, add a key to the
            # `_DatasetTracker` for the connections name, with a value of
            # the TaskDef in the appropriate field
            for inpt in iterConnections(connections, ("inputs", "prerequisiteInputs", "initInputs")):
                # Have to handle components in inputs.
                dataset_name, _, _ = inpt.name.partition(".")
                self._datasetDict.addConsumer(DatasetTypeName(dataset_name), taskDef)

            for output in iterConnections(connections, ("outputs",)):
                # Have to handle possible components in outputs.
                dataset_name, _, _ = output.name.partition(".")
                self._datasetDict.addProducer(DatasetTypeName(dataset_name), taskDef)

            # For each `Quantum` in the set of all `Quantum` for this task,
            # add a key to the `_DatasetTracker` that is a `DatasetRef` for one
            # of the individual datasets inside the `Quantum`, with a value of
            # a newly created QuantumNode to the appropriate input/output
            # field.
            for quantum in quantumSet:
                if quantum.dataId is not None:
                    if universe is None:
                        universe = quantum.dataId.universe
                    elif universe != quantum.dataId.universe:
                        raise RuntimeError(
                            "Mismatched dimension universes in QuantumGraph construction: "
                            f"{universe} != {quantum.dataId.universe}. "
                        )

                if _quantumToNodeId:
                    if (nodeId := _quantumToNodeId.get(quantum)) is None:
                        raise ValueError(
                            "If _quantuMToNodeNumber is not None, all quanta must have an "
                            "associated value in the mapping"
                        )
                else:
                    nodeId = generate_uuidv7()

                inits = quantum.initInputs.values()
                inputs = quantum.inputs.values()
                value = QuantumNode(quantum, taskDef, nodeId)
                self._taskToQuantumNode[taskDef].add(value)
                self._nodeIdMap[nodeId] = value

                bipartite_graph.add_node(value, bipartite=0)
                for dsRef in chain(inits, inputs):
                    # unfortunately, `Quantum` allows inits to be individual
                    # `DatasetRef`s or an Iterable of such, so there must
                    # be an instance check here
                    if isinstance(dsRef, Iterable):
                        for sub in dsRef:
                            bipartite_graph.add_node(sub.id.bytes, bipartite=1)
                            bipartite_graph.add_edge(sub.id.bytes, value)
                    else:
                        assert isinstance(dsRef, DatasetRef)
                        if dsRef.isComponent():
                            dsRef = dsRef.makeCompositeRef()
                        bipartite_graph.add_node(dsRef.id.bytes, bipartite=1)
                        bipartite_graph.add_edge(dsRef.id.bytes, value)
                for dsRef in chain.from_iterable(quantum.outputs.values()):
                    bipartite_graph.add_node(dsRef.id.bytes, bipartite=1)
                    bipartite_graph.add_edge(value, dsRef.id.bytes)

        # Dimension universe
        if universe is None:
            raise RuntimeError(
                "Dimension universe or at least one quantum with a data ID "
                "must be provided when constructing a QuantumGraph."
            )
        self._universe = universe

        # Make graph of quanta relations, by projecting out the dataset nodes
        # in the bipartite_graph, leaving just the quanta.
        self._connectedQuanta = nx.algorithms.bipartite.projected_graph(
            bipartite_graph, self._nodeIdMap.values()
        )
        self._count = len(self._connectedQuanta)

        # Graph of task relations, used in various methods
        self._taskGraph = self._datasetDict.makeNetworkXGraph()

        # convert default dict into a regular to prevent accidental key
        # insertion
        self._taskToQuantumNode = dict(self._taskToQuantumNode.items())

        self._initInputRefs: dict[str, list[DatasetRef]] = {}
        self._initOutputRefs: dict[str, list[DatasetRef]] = {}
        self._globalInitOutputRefs: list[DatasetRef] = []
        self._registryDatasetTypes: list[DatasetType] = []
        if initInputs is not None:
            self._initInputRefs = {taskDef.label: list(refs) for taskDef, refs in initInputs.items()}
        if initOutputs is not None:
            self._initOutputRefs = {taskDef.label: list(refs) for taskDef, refs in initOutputs.items()}
        if globalInitOutputs is not None:
            self._globalInitOutputRefs = list(globalInitOutputs)
        if registryDatasetTypes is not None:
            self._registryDatasetTypes = list(registryDatasetTypes)

        # PipelineGraph is current constructed on first use.
        # TODO DM-40442: use PipelineGraph instead of TaskDef
        # collections.
        self._pipeline_graph: PipelineGraph | None = None

    @property
    def pipeline_graph(self) -> PipelineGraph:
        """A graph representation of the tasks and dataset types in the quantum
        graph.
        """
        if self._pipeline_graph is None:
            # Construct into a temporary for strong exception safety.
            pipeline_graph = PipelineGraph()
            for task_def in self._taskToQuantumNode.keys():
                pipeline_graph.add_task(
                    task_def.label, task_def.taskClass, task_def.config, connections=task_def.connections
                )
            dataset_types = {dataset_type.name: dataset_type for dataset_type in self._registryDatasetTypes}
            pipeline_graph.resolve(dimensions=self._universe, dataset_types=dataset_types)
            self._pipeline_graph = pipeline_graph
        return self._pipeline_graph

    def get_task_quanta(self, label: str) -> Mapping[uuid.UUID, Quantum]:
        """Return the quanta associated with the given task label.

        Parameters
        ----------
        label : `str`
            Task label.

        Returns
        -------
        quanta : `~collections.abc.Mapping` [ uuid.UUID, `Quantum` ]
            Mapping from quantum ID to quantum.  Empty if ``label`` does not
            correspond to a task in this graph.
        """
        task_def = self.findTaskDefByLabel(label)
        if not task_def:
            return {}
        return {node.nodeId: node.quantum for node in self.getNodesForTask(task_def)}

    @property
    def taskGraph(self) -> nx.DiGraph:
        """A graph representing the relations between the tasks inside
        the quantum graph (`networkx.DiGraph`).
        """
        return self._taskGraph

    @property
    def graph(self) -> nx.DiGraph:
        """A graph representing the relations between all the `QuantumNode`
        objects (`networkx.DiGraph`).

        The graph should usually be iterated over, or passed to methods of this
        class, but sometimes direct access to the ``networkx`` object may be
        helpful.
        """
        return self._connectedQuanta

    @property
    def inputQuanta(self) -> Iterable[QuantumNode]:
        """The nodes that are inputs to the graph (iterable [`QuantumNode`]).

        These are the nodes that do not depend on any other nodes in the
        graph.
        """
        return (q for q, n in self._connectedQuanta.in_degree if n == 0)

    @property
    def outputQuanta(self) -> Iterable[QuantumNode]:
        """The nodes that are outputs of the graph (iterable [`QuantumNode`]).

        These are the nodes that have no nodes that depend on them in the
        graph.
        """
        return [q for q, n in self._connectedQuanta.out_degree if n == 0]

    @property
    def allDatasetTypes(self) -> tuple[DatasetTypeName, ...]:
        """All the data set type names that are present in the graph
        (`tuple` [`str`]).

        These types do not include global init-outputs.
        """
        return tuple(self._datasetDict.keys())

    @property
    def isConnected(self) -> bool:
        """Whether all of the nodes in the graph are connected, ignoring
        directionality of connections (`bool`).
        """
        return nx.is_weakly_connected(self._connectedQuanta)

    def getQuantumNodeByNodeId(self, nodeId: uuid.UUID) -> QuantumNode:
        """Lookup a `QuantumNode` from an id associated with the node.

        Parameters
        ----------
        nodeId : `NodeId`
            The number associated with a node.

        Returns
        -------
        node : `QuantumNode`
            The node corresponding with input number.

        Raises
        ------
        KeyError
            Raised if the requested nodeId is not in the graph.
        """
        return self._nodeIdMap[nodeId]

    def getQuantaForTask(self, taskDef: TaskDef) -> frozenset[Quantum]:
        """Return all the `~lsst.daf.butler.Quantum` associated with a
        `TaskDef`.

        Parameters
        ----------
        taskDef : `TaskDef`
            The `TaskDef` for which `~lsst.daf.butler.Quantum` are to be
            queried.

        Returns
        -------
        quanta : `frozenset` of `~lsst.daf.butler.Quantum`
            The `set` of `~lsst.daf.butler.Quantum` that is associated with the
            specified `TaskDef`.
        """
        return frozenset(node.quantum for node in self._taskToQuantumNode.get(taskDef, ()))

    def getNumberOfQuantaForTask(self, taskDef: TaskDef) -> int:
        """Return the number of `~lsst.daf.butler.Quantum` associated with
        a `TaskDef`.

        Parameters
        ----------
        taskDef : `TaskDef`
            The `TaskDef` for which `~lsst.daf.butler.Quantum` are to be
            queried.

        Returns
        -------
        count : `int`
            The number of `~lsst.daf.butler.Quantum` that are associated with
            the specified `TaskDef`.
        """
        return len(self._taskToQuantumNode.get(taskDef, ()))

    def getNodesForTask(self, taskDef: TaskDef) -> frozenset[QuantumNode]:
        r"""Return all the `QuantumNode`\s associated with a `TaskDef`.

        Parameters
        ----------
        taskDef : `TaskDef`
            The `TaskDef` for which `~lsst.daf.butler.Quantum` are to be
            queried.

        Returns
        -------
        nodes : `frozenset` [ `QuantumNode` ]
            A `frozenset` of `QuantumNode` that is associated with the
            specified `TaskDef`.
        """
        return frozenset(self._taskToQuantumNode[taskDef])

    def findTasksWithInput(self, datasetTypeName: DatasetTypeName) -> Iterable[TaskDef]:
        """Find all tasks that have the specified dataset type name as an
        input.

        Parameters
        ----------
        datasetTypeName : `str`
            A string representing the name of a dataset type to be queried,
            can also accept a `DatasetTypeName` which is a `~typing.NewType` of
            `str` for type safety in static type checking.

        Returns
        -------
        tasks : `~collections.abc.Iterable` [ `TaskDef` ]
            `TaskDef` objects that have the specified `DatasetTypeName` as an
            input, list will be empty if no tasks use specified
            `DatasetTypeName` as an input.

        Raises
        ------
        KeyError
            Raised if the `DatasetTypeName` is not part of the `QuantumGraph`.
        """
        return (c for c in self._datasetDict.getConsumers(datasetTypeName))

    def findTaskWithOutput(self, datasetTypeName: DatasetTypeName) -> TaskDef | None:
        """Find all tasks that have the specified dataset type name as an
        output.

        Parameters
        ----------
        datasetTypeName : `str`
            A string representing the name of a dataset type to be queried,
            can also accept a `DatasetTypeName` which is a `~typing.NewType` of
            `str` for type safety in static type checking.

        Returns
        -------
        result : `TaskDef` or `None`
            `TaskDef` that outputs `DatasetTypeName` as an output or `None` if
            none of the tasks produce this `DatasetTypeName`.

        Raises
        ------
        KeyError
            Raised if the `DatasetTypeName` is not part of the `QuantumGraph`.
        """
        return self._datasetDict.getProducer(datasetTypeName)

    def tasksWithDSType(self, datasetTypeName: DatasetTypeName) -> Iterable[TaskDef]:
        """Find all tasks that are associated with the specified dataset type
        name.

        Parameters
        ----------
        datasetTypeName : `str`
            A string representing the name of a dataset type to be queried,
            can also accept a `DatasetTypeName` which is a `~typing.NewType` of
            `str` for type safety in static type checking.

        Returns
        -------
        result : `~collections.abc.Iterable` [`TaskDef`]
            `TaskDef` objects that are associated with the specified
            `DatasetTypeName`.

        Raises
        ------
        KeyError
            Raised if the `DatasetTypeName` is not part of the `QuantumGraph`.
        """
        return self._datasetDict.getAll(datasetTypeName)

    def findTaskDefByName(self, taskName: str) -> list[TaskDef]:
        """Determine which `TaskDef` objects in this graph are associated
        with a `str` representing a task name (looks at the ``taskName``
        property of `TaskDef` objects).

        Returns a list of `TaskDef` objects as a `PipelineTask` may appear
        multiple times in a graph with different labels.

        Parameters
        ----------
        taskName : `str`
            Name of a task to search for.

        Returns
        -------
        result : `list` of `TaskDef`
            List of the `TaskDef` objects that have the name specified.
            Multiple values are returned in the case that a task is used
            multiple times with different labels.
        """
        results = []
        for task in self._taskToQuantumNode:
            split = task.taskName.split(".")
            if split[-1] == taskName:
                results.append(task)
        return results

    def findTaskDefByLabel(self, label: str) -> TaskDef | None:
        """Determine which `TaskDef` objects in this graph are associated
        with a `str` representing a tasks label.

        Parameters
        ----------
        label : `str`
            Name of a task to search for.

        Returns
        -------
        result : `TaskDef`
            `TaskDef` objects that has the specified label.
        """
        for task in self._taskToQuantumNode:
            if label == task.label:
                return task
        return None

    def findQuantaWithDSType(self, datasetTypeName: DatasetTypeName) -> set[Quantum]:
        r"""Return all the `~lsst.daf.butler.Quantum` that contain a specified
        `DatasetTypeName`.

        Parameters
        ----------
        datasetTypeName : `str`
            The name of the dataset type to search for as a string,
            can also accept a `DatasetTypeName` which is a `~typing.NewType` of
            `str` for type safety in static type checking.

        Returns
        -------
        result : `set` of `QuantumNode` objects
            A `set` of `QuantumNode`\s that contain specified
            `DatasetTypeName`.

        Raises
        ------
        KeyError
            Raised if the `DatasetTypeName` is not part of the `QuantumGraph`.
        """
        tasks = self._datasetDict.getAll(datasetTypeName)
        result: set[Quantum] = set()
        result = result.union(quantum for task in tasks for quantum in self.getQuantaForTask(task))
        return result

    def checkQuantumInGraph(self, quantum: Quantum) -> bool:
        """Check if specified quantum appears in the graph as part of a node.

        Parameters
        ----------
        quantum : `lsst.daf.butler.Quantum`
            The quantum to search for.

        Returns
        -------
        in_graph : `bool`
            The result of searching for the quantum.
        """
        return any(quantum == node.quantum for node in self)

    def writeDotGraph(self, output: str | io.BufferedIOBase) -> None:
        """Write out the graph as a dot graph.

        Parameters
        ----------
        output : `str` or `io.BufferedIOBase`
            Either a filesystem path to write to, or a file handle object.
        """
        write_dot(self._connectedQuanta, output)

    def subset(self: _T, nodes: QuantumNode | Iterable[QuantumNode]) -> _T:
        """Create a new graph object that contains the subset of the nodes
        specified as input. Node number is preserved.

        Parameters
        ----------
        nodes : `QuantumNode` or iterable of `QuantumNode`
            Nodes from which to create subset.

        Returns
        -------
        graph : instance of graph type
            An instance of the type from which the subset was created.
        """
        if not isinstance(nodes, Iterable):
            nodes = (nodes,)
        quantumSubgraph = self._connectedQuanta.subgraph(nodes).nodes
        quantumMap = defaultdict(set)

        dataset_type_names: set[str] = set()
        node: QuantumNode
        for node in quantumSubgraph:
            quantumMap[node.taskDef].add(node.quantum)
            dataset_type_names.update(
                dstype.name
                for dstype in chain(
                    node.quantum.inputs.keys(), node.quantum.outputs.keys(), node.quantum.initInputs.keys()
                )
            )

        # May need to trim dataset types from registryDatasetTypes.
        for taskDef in quantumMap:
            if refs := self.initOutputRefs(taskDef):
                dataset_type_names.update(ref.datasetType.name for ref in refs)
        dataset_type_names.update(ref.datasetType.name for ref in self._globalInitOutputRefs)
        registryDatasetTypes = [
            dstype for dstype in self._registryDatasetTypes if dstype.name in dataset_type_names
        ]

        # convert to standard dict to prevent accidental key insertion
        quantumDict: dict[TaskDef, set[Quantum]] = dict(quantumMap.items())
        # Create an empty graph, and then populate it with custom mapping
        newInst = type(self)({}, universe=self._universe)
        # TODO: Do we need to copy initInputs/initOutputs?
        newInst._buildGraphs(
            quantumDict,
            _quantumToNodeId={n.quantum: n.nodeId for n in nodes},
            _buildId=self._buildId,
            metadata=self._metadata,
            universe=self._universe,
            globalInitOutputs=self._globalInitOutputRefs,
            registryDatasetTypes=registryDatasetTypes,
        )
        return newInst

    def subsetToConnected(self: _T) -> tuple[_T, ...]:
        """Generate a list of subgraphs where each is connected.

        Returns
        -------
        result : `list` of `QuantumGraph`
            A list of graphs that are each connected.
        """
        return tuple(
            self.subset(connectedSet)
            for connectedSet in nx.weakly_connected_components(self._connectedQuanta)
        )

    def determineInputsToQuantumNode(self, node: QuantumNode) -> set[QuantumNode]:
        """Return a set of `QuantumNode` that are direct inputs to a specified
        node.

        Parameters
        ----------
        node : `QuantumNode`
            The node of the graph for which inputs are to be determined.

        Returns
        -------
        inputs : `set` of `QuantumNode`
            All the nodes that are direct inputs to specified node.
        """
        return set(self._connectedQuanta.predecessors(node))

    def determineOutputsOfQuantumNode(self, node: QuantumNode) -> set[QuantumNode]:
        """Return a set of `QuantumNode` that are direct outputs of a specified
        node.

        Parameters
        ----------
        node : `QuantumNode`
            The node of the graph for which outputs are to be determined.

        Returns
        -------
        outputs : `set` of `QuantumNode`
            All the nodes that are direct outputs to specified node.
        """
        return set(self._connectedQuanta.successors(node))

    def determineConnectionsOfQuantumNode(self: _T, node: QuantumNode) -> _T:
        """Return a graph of `QuantumNode` that are direct inputs and outputs
        of a specified node.

        Parameters
        ----------
        node : `QuantumNode`
            The node of the graph for which connected nodes are to be
            determined.

        Returns
        -------
        graph : graph of `QuantumNode`
            All the nodes that are directly connected to specified node.
        """
        nodes = self.determineInputsToQuantumNode(node).union(self.determineOutputsOfQuantumNode(node))
        nodes.add(node)
        return self.subset(nodes)

    def determineAncestorsOfQuantumNode(self: _T, node: QuantumNode) -> _T:
        """Return a graph of the specified node and all the ancestor nodes
        directly reachable by walking edges.

        Parameters
        ----------
        node : `QuantumNode`
            The node for which all ancestors are to be determined.

        Returns
        -------
        ancestors : graph of `QuantumNode`
            Graph of node and all of its ancestors.
        """
        predecessorNodes = nx.ancestors(self._connectedQuanta, node)
        predecessorNodes.add(node)
        return self.subset(predecessorNodes)

    def findCycle(self) -> list[tuple[QuantumNode, QuantumNode]]:
        """Check a graph for the presense of cycles and returns the edges of
        any cycles found, or an empty list if there is no cycle.

        Returns
        -------
        result : `list` of `tuple` of  [ `QuantumNode`, `QuantumNode` ]
            A list of any graph edges that form a cycle, or an empty list if
            there is no cycle. Empty list to so support if graph.find_cycle()
            syntax as an empty list is falsy.
        """
        try:
            return nx.find_cycle(self._connectedQuanta)
        except nx.NetworkXNoCycle:
            return []

    def saveUri(self, uri: ResourcePathExpression) -> None:
        """Save `QuantumGraph` to the specified URI.

        Parameters
        ----------
        uri : convertible to `~lsst.resources.ResourcePath`
            URI to where the graph should be saved.
        """
        path = ResourcePath(uri)
        match path.getExtension():
            case ".qgraph":
                buffer = self._buildSaveObject()
                path.write(buffer)  # type: ignore  # Ignore because bytearray is safe to use in place of bytes
            case ".qg":
                from ..quantum_graph import PredictedQuantumGraphComponents

                pqg = PredictedQuantumGraphComponents.from_old_quantum_graph(self)
                pqg.write(path)
            case ext:
                raise TypeError(f"Can currently only save a graph in .qgraph or .qg format, not {ext!r}.")

    @property
    def metadata(self) -> MappingProxyType[str, Any]:
        """Extra data carried with the graph (mapping [`str`] or `None`).

        The mapping is a dynamic view of this object's metadata. Values should
        be able to be serialized in JSON.
        """
        return MappingProxyType(self._metadata)

    def get_init_input_refs(self, task_label: str) -> list[DatasetRef]:
        """Return the DatasetRefs for the given task's init inputs.

        Parameters
        ----------
        task_label : `str`
            Label of the task.

        Returns
        -------
        refs : `list` [ `lsst.daf.butler.DatasetRef` ]
            Dataset references.  Guaranteed to be a new list, not internal
            state.
        """
        return list(self._initInputRefs.get(task_label, ()))

    def get_init_output_refs(self, task_label: str) -> list[DatasetRef]:
        """Return the DatasetRefs for the given task's init outputs.

        Parameters
        ----------
        task_label : `str`
            Label of the task.

        Returns
        -------
        refs : `list` [ `lsst.daf.butler.DatasetRef` ]
            Dataset references.  Guaranteed to be a new list, not internal
            state.
        """
        return list(self._initOutputRefs.get(task_label, ()))

    def initInputRefs(self, taskDef: TaskDef) -> list[DatasetRef] | None:
        """Return DatasetRefs for a given task InitInputs.

        Parameters
        ----------
        taskDef : `TaskDef`
            Task definition structure.

        Returns
        -------
        refs : `list` [ `~lsst.daf.butler.DatasetRef` ] or `None`
            DatasetRef for the task InitInput, can be `None`. This can return
            either resolved or non-resolved reference.
        """
        return self._initInputRefs.get(taskDef.label)

    def initOutputRefs(self, taskDef: TaskDef) -> list[DatasetRef] | None:
        """Return DatasetRefs for a given task InitOutputs.

        Parameters
        ----------
        taskDef : `TaskDef`
            Task definition structure.

        Returns
        -------
        refs : `list` [ `~lsst.daf.butler.DatasetRef` ] or `None`
            DatasetRefs for the task InitOutput, can be `None`. This can return
            either resolved or non-resolved reference. Resolved reference will
            match Quantum's initInputs if this is an intermediate dataset type.
        """
        return self._initOutputRefs.get(taskDef.label)

    def globalInitOutputRefs(self) -> list[DatasetRef]:
        """Return DatasetRefs for global InitOutputs.

        Returns
        -------
        refs : `list` [ `~lsst.daf.butler.DatasetRef` ]
            DatasetRefs for global InitOutputs.
        """
        return self._globalInitOutputRefs

    def registryDatasetTypes(self) -> list[DatasetType]:
        """Return dataset types used by this graph, their definitions match
        dataset types from registry.

        Returns
        -------
        refs : `list` [ `~lsst.daf.butler.DatasetType` ]
            Dataset types for this graph.
        """
        return self._registryDatasetTypes

    @classmethod
    def loadUri(
        cls,
        uri: ResourcePathExpression,
        universe: DimensionUniverse | None = None,
        nodes: Iterable[uuid.UUID | str] | None = None,
        graphID: BuildId | None = None,
        minimumVersion: int = 3,
    ) -> QuantumGraph:
        """Read `QuantumGraph` from a URI.

        Parameters
        ----------
        uri : convertible to `~lsst.resources.ResourcePath`
            URI from where to load the graph.
        universe : `~lsst.daf.butler.DimensionUniverse`, optional
            If `None` it is loaded from the `QuantumGraph`
            saved structure. If supplied, the
            `~lsst.daf.butler.DimensionUniverse` from the loaded `QuantumGraph`
            will be validated against the supplied argument for compatibility.
        nodes : `~collections.abc.Iterable` [ `uuid.UUID` | `str` ] or `None`
            UUIDs that correspond to nodes in the graph. If specified, only
            these nodes will be loaded. Defaults to None, in which case all
            nodes will be loaded.
        graphID : `str` or `None`
            If specified this ID is verified against the loaded graph prior to
            loading any Nodes. This defaults to None in which case no
            validation is done.
        minimumVersion : `int`
            Minimum version of a save file to load. Set to -1 to load all
            versions. Older versions may need to be loaded, and re-saved
            to upgrade them to the latest format before they can be used in
            production.

        Returns
        -------
        graph : `QuantumGraph`
            Resulting QuantumGraph instance.

        Raises
        ------
        TypeError
            Raised if file contains instance of a type other than
            `QuantumGraph`.
        ValueError
            Raised if one or more of the nodes requested is not in the
            `QuantumGraph` or if graphID parameter does not match the graph
            being loaded or if the supplied uri does not point at a valid
            `QuantumGraph` save file.
        RuntimeError
            Raise if Supplied `~lsst.daf.butler.DimensionUniverse` is not
            compatible with the `~lsst.daf.butler.DimensionUniverse` saved in
            the graph.
        """
        uri = ResourcePath(uri)
        match uri.getExtension():
            case ".qgraph":
                with LoadHelper(uri, minimumVersion, fullRead=(nodes is None)) as loader:
                    qgraph = loader.load(universe, nodes, graphID)
            case ".qg":
                from ..quantum_graph import PredictedQuantumGraphReader

                with PredictedQuantumGraphReader.open(uri, page_size=100000) as qgr:
                    quantum_ids = (
                        [uuid.UUID(q) if not isinstance(q, uuid.UUID) else q for q in nodes]
                        if nodes is not None
                        else None
                    )
                    qgr.read_execution_quanta(quantum_ids)
                    qgraph = qgr.finish().to_old_quantum_graph()
            case _:
                raise ValueError(f"Only know how to handle files saved as `.qgraph`, not {uri}")
        if not isinstance(qgraph, QuantumGraph):
            raise TypeError(f"QuantumGraph file {uri} contains unexpected object type: {type(qgraph)}")
        return qgraph

    @classmethod
    def readHeader(cls, uri: ResourcePathExpression, minimumVersion: int = 3) -> str | None:
        """Read the header of a `QuantumGraph` pointed to by the uri parameter
        and return it as a string.

        Parameters
        ----------
        uri : convertible to `~lsst.resources.ResourcePath`
            The location of the `QuantumGraph` to load. If the argument is a
            string, it must correspond to a valid
            `~lsst.resources.ResourcePath` path.
        minimumVersion : `int`
            Minimum version of a save file to load. Set to -1 to load all
            versions. Older versions may need to be loaded, and re-saved
            to upgrade them to the latest format before they can be used in
            production.

        Returns
        -------
        header : `str` or `None`
            The header associated with the specified `QuantumGraph` it there is
            one, else `None`.

        Raises
        ------
        ValueError
            Raised if the extension of the file specified by uri is not a
            `QuantumGraph` extension.
        """
        uri = ResourcePath(uri)
        if uri.getExtension() in {".qgraph"}:
            return LoadHelper(uri, minimumVersion).readHeader()
        else:
            raise ValueError("Only know how to handle files saved as `.qgraph`")

    def buildAndPrintHeader(self) -> None:
        """Create a header that would be used in a save of this object and
        prints it out to standard out.
        """
        _, header = self._buildSaveObject(returnHeader=True)
        print(json.dumps(header))

    def save(self, file: BinaryIO) -> None:
        """Save QuantumGraph to a file.

        Parameters
        ----------
        file : `io.BufferedIOBase`
            File to write data open in binary mode.
        """
        buffer = self._buildSaveObject()
        file.write(buffer)  # type: ignore # Ignore because bytearray is safe to use in place of bytes

    def _buildSaveObject(self, returnHeader: bool = False) -> bytearray | tuple[bytearray, dict]:
        thing = PersistenceContextVars()
        result = thing.run(self._buildSaveObjectImpl, returnHeader)
        return result

    def _buildSaveObjectImpl(self, returnHeader: bool = False) -> bytearray | tuple[bytearray, dict]:
        # make some containers
        jsonData: deque[bytes] = deque()
        # node map is a list because json does not accept mapping keys that
        # are not strings, so we store a list of key, value pairs that will
        # be converted to a mapping on load
        nodeMap = []
        taskDefMap = {}
        headerData: dict[str, Any] = {}

        # Store the QuantumGraph BuildId, this will allow validating BuildIds
        # at load time, prior to loading any QuantumNodes. Name chosen for
        # unlikely conflicts.
        headerData["GraphBuildID"] = self.graphID
        headerData["Metadata"] = self._metadata

        # Store the universe this graph was created with
        universeConfig = self._universe.dimensionConfig
        headerData["universe"] = universeConfig.toDict()

        # counter for the number of bytes processed thus far
        count = 0
        # serialize out the task Defs recording the start and end bytes of each
        # taskDef
        inverseLookup = self._datasetDict.inverse
        taskDef: TaskDef
        # sort by task label to ensure serialization happens in the same order
        for taskDef in self.taskGraph:
            # compressing has very little impact on saving or load time, but
            # a large impact on on disk size, so it is worth doing
            taskDescription: dict[str, Any] = {}
            # save the fully qualified name.
            taskDescription["taskName"] = get_full_type_name(taskDef.taskClass)
            # save the config as a text stream that will be un-persisted on the
            # other end
            stream = io.StringIO()
            taskDef.config.saveToStream(stream)
            taskDescription["config"] = stream.getvalue()
            taskDescription["label"] = taskDef.label
            if (refs := self._initInputRefs.get(taskDef.label)) is not None:
                taskDescription["initInputRefs"] = [ref.to_json() for ref in refs]
            if (refs := self._initOutputRefs.get(taskDef.label)) is not None:
                taskDescription["initOutputRefs"] = [ref.to_json() for ref in refs]

            inputs = []
            outputs = []

            # Determine the connection between all of tasks and save that in
            # the header as a list of connections and edges in each task
            # this will help in un-persisting, and possibly in a "quick view"
            # method that does not require everything to be un-persisted
            #
            # Typing returns can't be parameter dependent
            for connection in inverseLookup[taskDef]:  # type: ignore
                consumers = self._datasetDict.getConsumers(connection)
                producer = self._datasetDict.getProducer(connection)
                if taskDef in consumers:
                    # This checks if the task consumes the connection directly
                    # from the datastore or it is produced by another task
                    producerLabel = producer.label if producer is not None else "datastore"
                    inputs.append((producerLabel, connection))
                elif taskDef not in consumers and producer is taskDef:
                    # If there are no consumers for this tasks produced
                    # connection, the output will be said to be the datastore
                    # in which case the for loop will be a zero length loop
                    if not consumers:
                        outputs.append(("datastore", connection))
                    for td in consumers:
                        outputs.append((td.label, connection))

            # dump to json string, and encode that string to bytes and then
            # conpress those bytes
            dump = lzma.compress(json.dumps(taskDescription).encode(), preset=2)
            # record the sizing and relation information
            taskDefMap[taskDef.label] = {
                "bytes": (count, count + len(dump)),
                "inputs": inputs,
                "outputs": outputs,
            }
            count += len(dump)
            jsonData.append(dump)

        headerData["TaskDefs"] = taskDefMap

        # serialize the nodes, recording the start and end bytes of each node
        dimAccumulator = DimensionRecordsAccumulator()
        for node in self:
            # compressing has very little impact on saving or load time, but
            # a large impact on on disk size, so it is worth doing
            simpleNode = node.to_simple(accumulator=dimAccumulator)

            dump = lzma.compress(simpleNode.model_dump_json().encode(), preset=2)
            jsonData.append(dump)
            nodeMap.append(
                (
                    str(node.nodeId),
                    {
                        "bytes": (count, count + len(dump)),
                        "inputs": [str(n.nodeId) for n in self.determineInputsToQuantumNode(node)],
                        "outputs": [str(n.nodeId) for n in self.determineOutputsOfQuantumNode(node)],
                    },
                )
            )
            count += len(dump)

        headerData["DimensionRecords"] = {
            key: value.model_dump()
            for key, value in dimAccumulator.makeSerializedDimensionRecordMapping().items()
        }

        # need to serialize this as a series of key,value tuples because of
        # a limitation on how json cant do anything but strings as keys
        headerData["Nodes"] = nodeMap

        if self._globalInitOutputRefs:
            headerData["GlobalInitOutputRefs"] = [ref.to_json() for ref in self._globalInitOutputRefs]

        if self._registryDatasetTypes:
            headerData["RegistryDatasetTypes"] = [dstype.to_json() for dstype in self._registryDatasetTypes]

        # dump the headerData to json
        header_encode = lzma.compress(json.dumps(headerData).encode())

        # record the sizes as 2 unsigned long long numbers for a total of 16
        # bytes
        save_bytes = struct.pack(STRUCT_FMT_BASE, SAVE_VERSION)

        fmt_string = DESERIALIZER_MAP[SAVE_VERSION].FMT_STRING()
        map_lengths = struct.pack(fmt_string, len(header_encode))

        # write each component of the save out in a deterministic order
        buffer = bytearray()
        buffer.extend(MAGIC_BYTES)
        buffer.extend(save_bytes)
        buffer.extend(map_lengths)
        buffer.extend(header_encode)
        # Iterate over the length of jsonData, and for each element pop the
        # leftmost element off the deque and write it out. This is to save
        # memory, as the memory is added to the buffer object, it is removed
        # from from the container.
        #
        # Only this section needs to worry about memory pressure because
        # everything else written to the buffer prior to this data is
        # only on the order of kilobytes to low numbers of megabytes.
        while jsonData:
            buffer.extend(jsonData.popleft())
        if returnHeader:
            return buffer, headerData
        else:
            return buffer

    @classmethod
    def load(
        cls,
        file: BinaryIO,
        universe: DimensionUniverse | None = None,
        nodes: Iterable[uuid.UUID] | None = None,
        graphID: BuildId | None = None,
        minimumVersion: int = 3,
    ) -> QuantumGraph:
        """Read `QuantumGraph` from a file that was made by `save`.

        Parameters
        ----------
        file : `io.IO` of bytes
            File with data open in binary mode.
        universe : `~lsst.daf.butler.DimensionUniverse`, optional
            If `None` it is loaded from the `QuantumGraph`
            saved structure. If supplied, the
            `~lsst.daf.butler.DimensionUniverse` from the loaded `QuantumGraph`
            will be validated against the supplied argument for compatibility.
        nodes : `~collections.abc.Iterable` [`uuid.UUID`] or `None`
            UUIDs that correspond to nodes in the graph. If specified, only
            these nodes will be loaded. Defaults to None, in which case all
            nodes will be loaded.
        graphID : `str` or `None`
            If specified this ID is verified against the loaded graph prior to
            loading any Nodes. This defaults to None in which case no
            validation is done.
        minimumVersion : `int`
            Minimum version of a save file to load. Set to -1 to load all
            versions. Older versions may need to be loaded, and re-saved
            to upgrade them to the latest format before they can be used in
            production.

        Returns
        -------
        graph : `QuantumGraph`
            Resulting QuantumGraph instance.

        Raises
        ------
        TypeError
            Raised if data contains instance of a type other than
            `QuantumGraph`.
        ValueError
            Raised if one or more of the nodes requested is not in the
            `QuantumGraph` or if graphID parameter does not match the graph
            being loaded or if the supplied uri does not point at a valid
            `QuantumGraph` save file.
        """
        with LoadHelper(file, minimumVersion, fullRead=(nodes is None)) as loader:
            qgraph = loader.load(universe, nodes, graphID)
        if not isinstance(qgraph, QuantumGraph):
            raise TypeError(f"QuantumGraph file contains unexpected object type: {type(qgraph)}")
        return qgraph

    def iterTaskGraph(self) -> Generator[TaskDef]:
        """Iterate over the `taskGraph` attribute in topological order.

        Yields
        ------
        taskDef : `TaskDef`
            `TaskDef` objects in topological order.
        """
        yield from nx.topological_sort(self.taskGraph)

    def updateRun(self, run: str, *, metadata_key: str | None = None, update_graph_id: bool = False) -> None:
        """Change output run and dataset ID for each output dataset.

        Parameters
        ----------
        run : `str`
            New output run name.
        metadata_key : `str` or `None`
            Specifies matadata key corresponding to output run name to update
            with new run name. If `None` or if metadata is missing it is not
            updated. If metadata is present but key is missing, it will be
            added.
        update_graph_id : `bool`, optional
            If `True` then also update graph ID with a new unique value.
        """
        dataset_id_map: dict[DatasetId, DatasetId] = {}

        def _update_output_refs(
            refs: Iterable[DatasetRef], run: str, dataset_id_map: MutableMapping[DatasetId, DatasetId]
        ) -> Iterator[DatasetRef]:
            """Update a collection of `~lsst.daf.butler.DatasetRef` with new
            run and dataset IDs.
            """
            for ref in refs:
                new_ref = ref.replace(run=run)
                dataset_id_map[ref.id] = new_ref.id
                yield new_ref

        def _update_intermediate_refs(
            refs: Iterable[DatasetRef], run: str, dataset_id_map: Mapping[DatasetId, DatasetId]
        ) -> Iterator[DatasetRef]:
            """Update intermediate references with new run and IDs. Only the
            references that appear in ``dataset_id_map`` are updated, others
            are returned unchanged.
            """
            for ref in refs:
                if dataset_id := dataset_id_map.get(ref.id):
                    ref = ref.replace(run=run, id=dataset_id)
                yield ref

        # Replace quantum output refs first.
        for node in self._connectedQuanta:
            quantum = node.quantum
            outputs = {
                dataset_type: tuple(_update_output_refs(refs, run, dataset_id_map))
                for dataset_type, refs in quantum.outputs.items()
            }
            updated_quantum = Quantum(
                taskName=quantum.taskName,
                dataId=quantum.dataId,
                initInputs=quantum.initInputs,
                inputs=quantum.inputs,
                outputs=outputs,
                datastore_records=quantum.datastore_records,
            )
            node._replace_quantum(updated_quantum)

        self._initOutputRefs = {
            task_def: list(_update_output_refs(refs, run, dataset_id_map))
            for task_def, refs in self._initOutputRefs.items()
        }
        self._globalInitOutputRefs = list(
            _update_output_refs(self._globalInitOutputRefs, run, dataset_id_map)
        )

        # Update all intermediates from their matching outputs.
        for node in self._connectedQuanta:
            quantum = node.quantum
            inputs = {
                dataset_type: tuple(_update_intermediate_refs(refs, run, dataset_id_map))
                for dataset_type, refs in quantum.inputs.items()
            }
            initInputs = list(_update_intermediate_refs(quantum.initInputs.values(), run, dataset_id_map))

            updated_quantum = Quantum(
                taskName=quantum.taskName,
                dataId=quantum.dataId,
                initInputs=initInputs,
                inputs=inputs,
                outputs=quantum.outputs,
                datastore_records=quantum.datastore_records,
            )
            node._replace_quantum(updated_quantum)

        self._initInputRefs = {
            task_def: list(_update_intermediate_refs(refs, run, dataset_id_map))
            for task_def, refs in self._initInputRefs.items()
        }

        if update_graph_id:
            self._buildId = BuildId(f"{time.time()}-{os.getpid()}")

        # Update run if given.
        if metadata_key is not None:
            self._metadata[metadata_key] = run

    @property
    def graphID(self) -> BuildId:
        """The ID generated by the graph at construction time (`str`)."""
        return self._buildId

    @property
    def universe(self) -> DimensionUniverse:
        """Dimension universe associated with this graph
        (`~lsst.daf.butler.DimensionUniverse`).
        """
        return self._universe

    def __iter__(self) -> Generator[QuantumNode]:
        yield from nx.topological_sort(self._connectedQuanta)

    def __len__(self) -> int:
        return self._count

    def __contains__(self, node: QuantumNode) -> bool:
        return self._connectedQuanta.has_node(node)

    def __getstate__(self) -> dict:
        """Store a compact form of the graph as a list of graph nodes, and a
        tuple of task labels and task configs. The full graph can be
        reconstructed with this information, and it preserves the ordering of
        the graph nodes.
        """
        universe: DimensionUniverse | None = None
        for node in self:
            dId = node.quantum.dataId
            if dId is None:
                continue
            universe = dId.universe
        return {"reduced": self._buildSaveObject(), "graphId": self._buildId, "universe": universe}

    def __setstate__(self, state: dict) -> None:
        """Reconstructs the state of the graph from the information persisted
        in getstate.
        """
        buffer = io.BytesIO(state["reduced"])
        with LoadHelper(buffer, minimumVersion=3) as loader:
            qgraph = loader.load(state["universe"], graphID=state["graphId"])

        self._metadata = qgraph._metadata
        self._buildId = qgraph._buildId
        self._datasetDict = qgraph._datasetDict
        self._nodeIdMap = qgraph._nodeIdMap
        self._count = len(qgraph)
        self._taskToQuantumNode = qgraph._taskToQuantumNode
        self._taskGraph = qgraph._taskGraph
        self._connectedQuanta = qgraph._connectedQuanta
        self._initInputRefs = qgraph._initInputRefs
        self._initOutputRefs = qgraph._initOutputRefs

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QuantumGraph):
            return False
        if len(self) != len(other):
            return False
        for node in self:
            if node not in other:
                return False
            if self.determineInputsToQuantumNode(node) != other.determineInputsToQuantumNode(node):
                return False
            if self.determineOutputsOfQuantumNode(node) != other.determineOutputsOfQuantumNode(node):
                return False
        if set(self.allDatasetTypes) != set(other.allDatasetTypes):
            return False
        return set(self.taskGraph) == set(other.taskGraph)

    def getSummary(self) -> QgraphSummary:
        """Create summary of graph.

        Returns
        -------
        summary : `QgraphSummary`
            Summary of QuantumGraph.
        """
        inCollection = self.metadata.get("input", None)
        if isinstance(inCollection, str):
            inCollection = [inCollection]
        summary = QgraphSummary(
            graphID=self.graphID,
            cmdLine=self.metadata.get("full_command", None),
            creationUTC=self.metadata.get("time", None),
            inputCollection=inCollection,
            outputCollection=self.metadata.get("output", None),
            outputRun=self.metadata.get("output_run", None),
        )
        for q in self:
            qts = summary.qgraphTaskSummaries.setdefault(
                q.taskDef.label, QgraphTaskSummary(taskLabel=q.taskDef.label)
            )
            qts.numQuanta += 1

            for k in q.quantum.inputs.keys():
                qts.numInputs[k.name] += 1

            for k in q.quantum.outputs.keys():
                qts.numOutputs[k.name] += 1

        return summary

    def make_init_qbb(
        self,
        butler_config: Config | ResourcePathExpression,
        *,
        config_search_paths: Iterable[str] | None = None,
    ) -> QuantumBackedButler:
        """Construct an quantum-backed butler suitable for reading and writing
        init input and init output datasets, respectively.

        This requires the full graph to have been loaded.

        Parameters
        ----------
        butler_config : `~lsst.daf.butler.Config` or \
                `~lsst.resources.ResourcePathExpression`
            A butler repository root, configuration filename, or configuration
            instance.
        config_search_paths : `~collections.abc.Iterable` [ `str` ], optional
            Additional search paths for butler configuration.

        Returns
        -------
        qbb : `~lsst.daf.butler.QuantumBackedButler`
            A limited butler that can ``get`` init-input datasets and ``put``
            init-output datasets.
        """
        universe = self.universe
        # Collect all init input/output dataset IDs.
        predicted_inputs: set[DatasetId] = set()
        predicted_outputs: set[DatasetId] = set()
        pipeline_graph = self.pipeline_graph
        for task_label in pipeline_graph.tasks:
            predicted_inputs.update(ref.id for ref in self.get_init_input_refs(task_label))
            predicted_outputs.update(ref.id for ref in self.get_init_output_refs(task_label))
        predicted_outputs.update(ref.id for ref in self.globalInitOutputRefs())
        # remove intermediates from inputs
        predicted_inputs -= predicted_outputs
        # Very inefficient way to extract datastore records from quantum graph,
        # we have to scan all quanta and look at their datastore records.
        datastore_records: dict[str, DatastoreRecordData] = {}
        for quantum_node in self:
            for store_name, records in quantum_node.quantum.datastore_records.items():
                subset = records.subset(predicted_inputs)
                if subset is not None:
                    datastore_records.setdefault(store_name, DatastoreRecordData()).update(subset)

        dataset_types = {dstype.name: dstype for dstype in self.registryDatasetTypes()}
        # Make butler from everything.
        return QuantumBackedButler.from_predicted(
            config=butler_config,
            predicted_inputs=predicted_inputs,
            predicted_outputs=predicted_outputs,
            dimensions=universe,
            datastore_records=datastore_records,
            search_paths=list(config_search_paths) if config_search_paths is not None else None,
            dataset_types=dataset_types,
        )

    def write_init_outputs(self, butler: LimitedButler, skip_existing: bool = True) -> None:
        """Write the init-output datasets for all tasks in the quantum graph.

        Parameters
        ----------
        butler : `lsst.daf.butler.LimitedButler`
            A limited butler data repository client.
        skip_existing : `bool`, optional
            If `True` (default) ignore init-outputs that already exist.  If
            `False`, raise.

        Raises
        ------
        lsst.daf.butler.registry.ConflictingDefinitionError
            Raised if an init-output dataset already exists and
            ``skip_existing=False``.
        """
        # Extract init-input and init-output refs from the QG.
        input_refs: dict[str, DatasetRef] = {}
        output_refs: dict[str, DatasetRef] = {}
        for task_node in self.pipeline_graph.tasks.values():
            input_refs.update(
                {ref.datasetType.name: ref for ref in self.get_init_input_refs(task_node.label)}
            )
            output_refs.update(
                {
                    ref.datasetType.name: ref
                    for ref in self.get_init_output_refs(task_node.label)
                    if ref.datasetType.name != task_node.init.config_output.dataset_type_name
                }
            )
        for ref, is_stored in butler.stored_many(output_refs.values()).items():
            if is_stored:
                if not skip_existing:
                    raise ConflictingDefinitionError(f"Init-output dataset {ref} already exists.")
                # We'll `put` whatever's left in output_refs at the end.
                del output_refs[ref.datasetType.name]
        # Instantiate tasks, reading overall init-inputs and gathering
        # init-output in-memory objects.
        init_outputs: list[tuple[Any, DatasetType]] = []
        self.pipeline_graph.instantiate_tasks(
            get_init_input=lambda dataset_type: butler.get(
                input_refs[dataset_type.name].overrideStorageClass(dataset_type.storageClass)
            ),
            init_outputs=init_outputs,
        )
        # Write init-outputs that weren't already present.
        for obj, dataset_type in init_outputs:
            if new_ref := output_refs.get(dataset_type.name):
                assert new_ref.datasetType.storageClass_name == dataset_type.storageClass_name, (
                    "QG init refs should use task connection storage classes."
                )
                butler.put(obj, new_ref)

    def write_configs(self, butler: LimitedButler, compare_existing: bool = True) -> None:
        """Write the config datasets for all tasks in the quantum graph.

        Parameters
        ----------
        butler : `lsst.daf.butler.LimitedButler`
            A limited butler data repository client.
        compare_existing : `bool`, optional
            If `True` check configs that already exist for consistency.  If
            `False`, always raise if configs already exist.

        Raises
        ------
        lsst.daf.butler.registry.ConflictingDefinitionError
            Raised if an config dataset already exists and
            ``compare_existing=False``, or if the existing config is not
            consistent with the config in the quantum graph.
        """
        to_put: list[tuple[PipelineTaskConfig, DatasetRef]] = []
        for task_node in self.pipeline_graph.tasks.values():
            dataset_type_name = task_node.init.config_output.dataset_type_name
            (ref,) = [  # noqa: UP027
                ref
                for ref in self.get_init_output_refs(task_node.label)
                if ref.datasetType.name == dataset_type_name
            ]
            try:
                old_config = butler.get(ref)
            except (LookupError, FileNotFoundError):
                old_config = None
            if old_config is not None:
                if not compare_existing:
                    raise ConflictingDefinitionError(f"Config dataset {ref} already exists.")
                if not task_node.config.compare(old_config, shortcut=False, output=log_config_mismatch):
                    raise ConflictingDefinitionError(
                        f"Config does not match existing task config {dataset_type_name!r} in "
                        "butler; tasks configurations must be consistent within the same run collection."
                    )
            else:
                to_put.append((task_node.config, ref))
        # We do writes at the end to minimize the mess we leave behind when we
        # raise an exception.
        for config, ref in to_put:
            butler.put(config, ref)

    def write_packages(self, butler: LimitedButler, compare_existing: bool = True) -> None:
        """Write the 'packages' dataset for the currently-active software
        versions.

        Parameters
        ----------
        butler : `lsst.daf.butler.LimitedButler`
            A limited butler data repository client.
        compare_existing : `bool`, optional
            If `True` check packages that already exist for consistency.  If
            `False`, always raise if the packages dataset already exists.

        Raises
        ------
        lsst.daf.butler.registry.ConflictingDefinitionError
            Raised if the packages dataset already exists and is not consistent
            with the current packages.
        """
        new_packages = Packages.fromSystem()
        (ref,) = self.globalInitOutputRefs()
        try:
            packages = butler.get(ref)
        except (LookupError, FileNotFoundError):
            packages = None
        if packages is not None:
            if not compare_existing:
                raise ConflictingDefinitionError(f"Packages dataset {ref} already exists.")
            if compare_packages(packages, new_packages):
                # have to remove existing dataset first; butler has no
                # replace option.
                butler.pruneDatasets([ref], unstore=True, purge=True)
                butler.put(packages, ref)
        else:
            butler.put(new_packages, ref)

    def init_output_run(self, butler: LimitedButler, existing: bool = True) -> None:
        """Initialize a new output RUN collection by writing init-output
        datasets (including configs and packages).

        Parameters
        ----------
        butler : `lsst.daf.butler.LimitedButler`
            A limited butler data repository client.
        existing : `bool`, optional
            If `True` check or ignore outputs that already exist.  If
            `False`, always raise if an output dataset already exists.

        Raises
        ------
        lsst.daf.butler.registry.ConflictingDefinitionError
            Raised if there are existing init output datasets, and either
            ``existing=False`` or their contents are not compatible with this
            graph.
        """
        self.write_configs(butler, compare_existing=existing)
        self.write_packages(butler, compare_existing=existing)
        self.write_init_outputs(butler, skip_existing=existing)

    def get_refs(
        self,
        *,
        include_init_inputs: bool = False,
        include_inputs: bool = False,
        include_intermediates: bool | None = None,
        include_init_outputs: bool = False,
        include_outputs: bool = False,
        conform_outputs: bool = True,
    ) -> tuple[set[DatasetRef], dict[str, DatastoreRecordData]]:
        """Get the requested dataset refs from the graph.

        Parameters
        ----------
        include_init_inputs : `bool`, optional
            Include init inputs.
        include_inputs : `bool`, optional
            Include inputs.
        include_intermediates : `bool` or `None`, optional
            If `None`, no special handling for intermediates is performed.
            If `True` intermediates are calculated even if other flags
            do not request datasets. If `False` intermediates will be removed
            from any results.
        include_init_outputs : `bool`, optional
            Include init outpus.
        include_outputs : `bool`, optional
            Include outputs.
        conform_outputs : `bool`, optional
            Whether any outputs found should have their dataset types conformed
            with the registry dataset types.

        Returns
        -------
        refs : `set` [ `lsst.daf.butler.DatasetRef` ]
            The requested dataset refs found in the graph.
        datastore_records : `dict` [ `str`, \
                `lsst.daf.butler.datastore.record_data.DatastoreRecordData` ]
            Any datastore records found.

        Notes
        -----
        Conforming and requesting inputs and outputs can result in the same
        dataset appearing in the results twice with differing storage classes.
        """
        datastore_records: dict[str, DatastoreRecordData] = {}
        init_input_refs: set[DatasetRef] = set()
        init_output_refs: set[DatasetRef] = set(self.globalInitOutputRefs())

        if include_intermediates is True:
            # Need to enable inputs and outputs even if not explicitly
            # requested.
            request_include_init_inputs = True
            request_include_inputs = True
            request_include_init_outputs = True
            request_include_outputs = True
        else:
            request_include_init_inputs = include_init_inputs
            request_include_inputs = include_inputs
            request_include_init_outputs = include_init_outputs
            request_include_outputs = include_outputs

        if request_include_init_inputs or request_include_init_outputs:
            for task_def in self.iterTaskGraph():
                if request_include_init_inputs:
                    if in_refs := self.initInputRefs(task_def):
                        init_input_refs.update(in_refs)
                if request_include_init_outputs:
                    if out_refs := self.initOutputRefs(task_def):
                        init_output_refs.update(out_refs)

        input_refs: set[DatasetRef] = set()
        output_refs: set[DatasetRef] = set()

        for qnode in self:
            if request_include_inputs:
                for other_refs in qnode.quantum.inputs.values():
                    input_refs.update(other_refs)
                # Inputs can come with datastore records.
                for store_name, records in qnode.quantum.datastore_records.items():
                    datastore_records.setdefault(store_name, DatastoreRecordData()).update(records)
            if request_include_outputs:
                for other_refs in qnode.quantum.outputs.values():
                    output_refs.update(other_refs)

        # Intermediates are the intersection of inputs and outputs. Must do
        # this analysis before conforming since dataset type changes will
        # change set membership.
        inter_msg = ""
        intermediates = set()
        if include_intermediates is not None:
            intermediates = (input_refs | init_input_refs) & (output_refs | init_output_refs)

        if include_intermediates is False:
            # Remove intermediates from results.
            init_input_refs -= intermediates
            input_refs -= intermediates
            init_output_refs -= intermediates
            output_refs -= intermediates
            inter_msg = f"; Intermediates removed: {len(intermediates)}"
            intermediates = set()
        elif include_intermediates is True:
            # Do not mention intermediates if all the input/output flags
            # would have resulted in them anyhow.
            if (
                (request_include_init_inputs is not include_init_inputs)
                or (request_include_inputs is not include_inputs)
                or (request_include_init_outputs is not include_init_outputs)
                or (request_include_outputs is not include_outputs)
            ):
                inter_msg = f"; including intermediates: {len(intermediates)}"

        # Assign intermediates to the relevant category.
        if not include_init_inputs:
            init_input_refs &= intermediates
        if not include_inputs:
            input_refs &= intermediates
        if not include_init_outputs:
            init_output_refs &= intermediates
        if not include_outputs:
            output_refs &= intermediates

        # Conforming can result in an input ref and an output ref appearing
        # in the returned results that are identical apart from storage class.
        if conform_outputs:
            # Get data repository definitions from the QuantumGraph; these can
            # have different storage classes than those in the quanta.
            dataset_types = {dstype.name: dstype for dstype in self.registryDatasetTypes()}

            def _update_ref(ref: DatasetRef) -> DatasetRef:
                internal_dataset_type = dataset_types.get(ref.datasetType.name, ref.datasetType)
                if internal_dataset_type.storageClass_name != ref.datasetType.storageClass_name:
                    ref = ref.replace(storage_class=internal_dataset_type.storageClass_name)
                return ref

            # Convert output_refs to the data repository storage classes, too.
            output_refs = {_update_ref(ref) for ref in output_refs}
            init_output_refs = {_update_ref(ref) for ref in init_output_refs}

        _LOG.verbose(
            "Found the following datasets. InitInputs: %d; Inputs: %d; InitOutputs: %s; Outputs: %d%s",
            len(init_input_refs),
            len(input_refs),
            len(init_output_refs),
            len(output_refs),
            inter_msg,
        )

        refs = input_refs | init_input_refs | init_output_refs | output_refs
        return refs, datastore_records



# This file is part of pipe_Base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ["LogCapture"]

import dataclasses
import logging
import uuid
from collections.abc import Iterator
from contextlib import contextmanager
from logging import FileHandler

import pydantic

from lsst.daf.butler import Butler, LimitedButler, Quantum
from lsst.daf.butler._rubin.temporary_for_ingest import TemporaryForIngest
from lsst.daf.butler.logging import (
    ButlerLogRecord,
    ButlerLogRecordHandler,
    ButlerLogRecords,
    ButlerMDC,
    JsonLogFormatter,
)

from ._status import ExceptionInfo, InvalidQuantumError
from ._task_metadata import TaskMetadata
from .automatic_connection_constants import METADATA_OUTPUT_TEMPLATE
from .pipeline_graph import TaskNode

_LOG = logging.getLogger(__name__)


class _ExecutionLogRecordsExtra(pydantic.BaseModel):
    """Extra information about a quantum's execution stored with logs.

    This middleware-private model includes information that is not directly
    available via any public interface, as it is used exclusively for
    provenance extraction and then made available through the provenance
    quantum graph.
    """

    exception: ExceptionInfo | None = None
    """Exception information for this quantum, if it failed.
    """

    metadata: TaskMetadata | None = None
    """Metadata for this quantum, if it failed.

    Metadata datasets are written if and only if a quantum succeeds, but we
    still want to capture metadata from failed attempts, so we store it in the
    log dataset.  This field is always `None` when the quantum succeeds,
    because in that case the metadata is already stored separately.
    """

    previous_process_quanta: list[uuid.UUID] = pydantic.Field(default_factory=list)
    """The IDs of other quanta previously executed in the same process as this
    one.
    """

    logs: list[ButlerLogRecord] = pydantic.Field(default_factory=list)
    """Logs for this attempt.

    This is always empty for the most recent attempt, because that stores logs
    in the main section of the butler log records.
    """

    previous_attempts: list[_ExecutionLogRecordsExtra] = pydantic.Field(default_factory=list)
    """Information about previous attempts to run this task within the same
    `~lsst.daf.butler.CollectionType.RUN` collection.

    This is always empty for any attempt other than the most recent one,
    as all previous attempts are flattened into one list.
    """

    def attach_previous_attempt(self, log_records: ButlerLogRecords) -> None:
        """Attach logs from a previous attempt to this struct.

        Parameters
        ----------
        log_records : `lsst.daf.butler.ButlerLogRecords`
            Logs from a past attempt to run a quantum.
        """
        previous = self.model_validate(log_records.extra)
        previous.logs.extend(log_records)
        self.previous_attempts.extend(previous.previous_attempts)
        self.previous_attempts.append(previous)
        previous.previous_attempts.clear()


@dataclasses.dataclass
class _LogCaptureContext:
    """Controls for log capture returned by the `LogCapture.capture_logging`
    context manager.
    """

    store: bool = True
    """Whether to store logs at all."""

    extra: _ExecutionLogRecordsExtra = dataclasses.field(default_factory=_ExecutionLogRecordsExtra)
    """Extra information about the quantum's execution to store for provenance
    extraction.
    """


class LogCapture:
    """Class handling capture of logging messages and their export to butler.

    Parameters
    ----------
    butler : `~lsst.daf.butler.LimitedButler`
        Data butler with limited API.
    full_butler : `~lsst.daf.butler.Butler` or `None`
        Data butler with full API, or `None` if full Butler is not available.
        If not none, then this must be the same instance as ``butler``.
    """

    stream_json_logs = True
    """If True each log record is written to a temporary file and ingested
    when quantum completes. If False the records are accumulated in memory
    and stored in butler on quantum completion. If full butler is not available
    then temporary file is not used."""

    def __init__(
        self,
        butler: LimitedButler,
        full_butler: Butler | None,
    ):
        self.butler = butler
        self.full_butler = full_butler

    @classmethod
    def from_limited(cls, butler: LimitedButler) -> LogCapture:
        return cls(butler, None)

    @classmethod
    def from_full(cls, butler: Butler) -> LogCapture:
        return cls(butler, butler)

    @contextmanager
    def capture_logging(
        self, task_node: TaskNode, /, quantum: Quantum, records: ButlerLogRecords | None = None
    ) -> Iterator[_LogCaptureContext]:
        """Configure logging system to capture logs for execution of this task.

        Parameters
        ----------
        task_node : `~lsst.pipe.base.pipeline_graph.TaskNode`
            The task definition.
        quantum : `~lsst.daf.butler.Quantum`
            Single Quantum instance.
        records : `lsst.daf.butler.logging.ButlerLogRecords`, optional
            Log record container to append to and save.  If provided, streaming
            mode is disabled (since we'll be saving logs in memory anyway).

        Notes
        -----
        Expected to be used as a context manager to ensure that logging
        records are inserted into the butler once the quantum has been
        executed:

        .. code-block:: py

           with self.capture_logging(task_node, quantum):
               # Run quantum and capture logs.

        Ths method can also setup logging to attach task- or
        quantum-specific information to log messages. Potentially this can
        take into account some info from task configuration as well.
        """
        # include quantum dataId and task label into MDC
        mdc = {"LABEL": task_node.label, "RUN": ""}
        if quantum.dataId:
            mdc["LABEL"] += f":{quantum.dataId}"

        metadata_ref = quantum.outputs[METADATA_OUTPUT_TEMPLATE.format(label=task_node.label)][0]
        mdc["RUN"] = metadata_ref.run

        ctx = _LogCaptureContext()
        log_dataset_name = (
            task_node.log_output.dataset_type_name if task_node.log_output is not None else None
        )

        # Add a handler to the root logger to capture execution log output.
        if log_dataset_name is not None:
            try:
                [ref] = quantum.outputs[log_dataset_name]
            except LookupError as exc:
                raise InvalidQuantumError(
                    f"Quantum outputs is missing log output dataset type {log_dataset_name};"
                    " this could happen due to inconsistent options between QuantumGraph generation"
                    " and execution"
                ) from exc
            # Either accumulate into ButlerLogRecords or stream JSON records to
            # file and ingest that (ingest is possible only with full butler).
            if self.stream_json_logs and self.full_butler is not None and records is None:
                with TemporaryForIngest(self.full_butler, ref) as temporary:
                    log_handler_file = FileHandler(temporary.ospath)
                    log_handler_file.setFormatter(JsonLogFormatter())
                    logging.getLogger().addHandler(log_handler_file)

                    try:
                        with ButlerMDC.set_mdc(mdc):
                            yield ctx
                    finally:
                        # Ensure that the logs are stored in butler.
                        logging.getLogger().removeHandler(log_handler_file)
                        log_handler_file.close()
                        if ctx.extra:
                            with open(temporary.ospath, "a") as log_stream:
                                ButlerLogRecords.write_streaming_extra(
                                    log_stream,
                                    ctx.extra.model_dump_json(exclude_unset=True, exclude_defaults=True),
                                )
                        if ctx.store:
                            temporary.ingest()

            else:
                log_handler_memory = ButlerLogRecordHandler(records)
                logging.getLogger().addHandler(log_handler_memory)

                try:
                    with ButlerMDC.set_mdc(mdc):
                        yield ctx
                except:
                    raise
                else:
                    # If the quantum succeeded, we don't need to save the
                    # metadata in the logs, because we'll have saved them in
                    # the metadata.
                    ctx.extra.metadata = None
                finally:
                    log_handler_memory.records.extra = ctx.extra.model_dump()
                    # Ensure that the logs are stored in butler.
                    logging.getLogger().removeHandler(log_handler_memory)
                    if ctx.store:
                        self._store_log_records(quantum, log_dataset_name, log_handler_memory)

        else:
            with ButlerMDC.set_mdc(mdc):
                yield ctx

    def _store_log_records(
        self, quantum: Quantum, dataset_type: str, log_handler: ButlerLogRecordHandler
    ) -> None:
        # DatasetRef has to be in the Quantum outputs, can lookup by name.
        try:
            [ref] = quantum.outputs[dataset_type]
        except LookupError as exc:
            raise InvalidQuantumError(
                f"Quantum outputs is missing log output dataset type {dataset_type};"
                " this could happen due to inconsistent options between QuantumGraph generation"
                " and execution"
            ) from exc

        self.butler.put(log_handler.records, ref)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from lsst.pipe.base.connectionTypes import BaseInput, Output

__all__ = (
    "DynamicConnectionConfig",
    "DynamicTestPipelineTask",
    "DynamicTestPipelineTaskConfig",
    "ForcedFailure",
    "MockAlgorithmError",
    "MockPipelineTask",
    "MockPipelineTaskConfig",
    "mock_pipeline_graph",
)

import dataclasses
import logging
import signal
import time
from collections.abc import Collection, Iterable, Mapping
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

from astropy.units import Quantity

from lsst.daf.butler import DataCoordinate, DatasetRef, DeferredDatasetHandle, Quantum, SerializedDatasetType
from lsst.pex.config import Config, ConfigDictField, ConfigurableField, Field, ListField
from lsst.utils.doImport import doImportType
from lsst.utils.introspection import get_full_type_name
from lsst.utils.iteration import ensure_iterable

from ... import connectionTypes as cT
from ..._status import AlgorithmError, AnnotatedPartialOutputsError
from ...automatic_connection_constants import METADATA_OUTPUT_CONNECTION_NAME, METADATA_OUTPUT_STORAGE_CLASS
from ...config import PipelineTaskConfig
from ...connections import InputQuantizedConnection, OutputQuantizedConnection, PipelineTaskConnections
from ...pipeline_graph import PipelineGraph
from ...pipelineTask import PipelineTask
from ._data_id_match import DataIdMatch
from ._storage_class import (
    ConvertedUnmockedDataset,
    MockDataset,
    MockDatasetQuantum,
    MockStorageClass,
    get_mock_name,
)

_LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..._quantumContext import QuantumContext


_T = TypeVar("_T", bound=cT.BaseConnection)


@dataclasses.dataclass
class ForcedFailure:
    """Information about an exception that should be raised by one or more
    quanta.
    """

    condition: str
    """Butler expression-language string that matches the data IDs that should
    raise.
    """

    exception_type: type[BaseException] | None = None
    """The type of exception to raise."""

    memory_required: Quantity | None = None
    """If not `None`, this failure simulates an out-of-memory failure by
    raising only if this value exceeds `ExecutionResources.max_mem`.
    """

    def set_config(self, config: MockPipelineTaskConfig) -> None:
        config.fail_condition = self.condition
        if self.exception_type:
            config.fail_exception = get_full_type_name(self.exception_type)
        config.memory_required = self.memory_required


class MockAlgorithmError(AlgorithmError):
    """A subclass of `..AlgorithmError` chained to
    `..AnnotatedPartialOutputsError` when the latter is configured to be raised
    by `MockPipelineTask`.
    """

    @property
    def metadata(self) -> dict[str, int]:
        return {"badness": 12}


def mock_pipeline_graph(
    original_graph: PipelineGraph,
    unmocked_dataset_types: Iterable[str] = (),
    force_failures: Mapping[str, ForcedFailure] | None = None,
) -> PipelineGraph:
    """Create mocks for a full pipeline graph.

    Parameters
    ----------
    original_graph : `~..pipeline_graph.PipelineGraph`
        Original tasks and configuration to mock.
    unmocked_dataset_types : `~collections.abc.Iterable` [ `str` ], optional
        Names of overall-input dataset types that should not be replaced with
        mocks.  "Automatic" datasets written by the execution framework such
        as configs, logs, and metadata are implicitly included.
    force_failures : `~collections.abc.Mapping` [ `str`, `ForcedFailure` ]
        Mapping from original task label to information about an exception one
        or more quanta for this task should raise.

    Returns
    -------
    mocked : `~..pipeline_graph.PipelineGraph`
        Pipeline graph using `MockPipelineTask` configurations that target the
        original tasks.  Never resolved.
    """
    unmocked_dataset_types = list(unmocked_dataset_types)
    if force_failures is None:
        force_failures = {}
    result = PipelineGraph(description=original_graph.description)
    for task_node in original_graph.tasks.values():
        unmocked_dataset_types.append(task_node.init.config_output.dataset_type_name)
        if task_node.log_output is not None:
            unmocked_dataset_types.append(task_node.log_output.dataset_type_name)
        unmocked_dataset_types.append(task_node.metadata_output.dataset_type_name)
    for original_task_node in original_graph.tasks.values():
        config = MockPipelineTaskConfig()
        config.original.retarget(original_task_node.task_class)
        config.original = original_task_node.config
        config.unmocked_dataset_types.extend(unmocked_dataset_types)
        if original_task_node.label in force_failures:
            force_failures[original_task_node.label].set_config(config)
        result.add_task(get_mock_name(original_task_node.label), MockPipelineTask, config=config)
    return result


class BaseTestPipelineTaskConnections(PipelineTaskConnections, dimensions=()):
    pass


class BaseTestPipelineTaskConfig(PipelineTaskConfig, pipelineConnections=BaseTestPipelineTaskConnections):
    fail_condition = Field[str](
        dtype=str,
        default="",
        doc=(
            "Condition on Data ID to raise an exception. String expression which includes attributes of "
            "quantum data ID using a syntax of daf_butler user expressions (e.g. 'visit = 123')."
        ),
    )

    fail_exception = Field[str](
        dtype=str,
        default="builtins.ValueError",
        doc=(
            "Class name of the exception to raise when fail condition is triggered. Can be "
            "'lsst.pipe.base.NoWorkFound' to specify non-failure exception."
        ),
    )

    fail_signal = Field[int](dtype=int, optional=True, doc="Signal to raise instead of an exception.")

    memory_required = Field[str](
        dtype=str,
        default=None,
        optional=True,
        doc=(
            "If not None, simulate an out-of-memory failure by raising only if ExecutionResource.max_mem "
            "exceeds this value.  This string should include units as parsed by astropy.units.Quantity "
            "(e.g. '4GB')."
        ),
    )

    sleep = Field[float](
        dtype=float,
        default=0.0,
        doc="Time to sleep (seconds) before mock execution reading inputs or failing.",
    )

    int_value = Field[int](
        "Arbitrary integer value to write into mock output datasets", dtype=int, optional=True, default=None
    )
    str_value = Field[str](
        "Arbitrary string value to write into mock output datasets", dtype=str, optional=True, default=None
    )

    def data_id_match(self) -> DataIdMatch | None:
        if not self.fail_condition:
            return None
        return DataIdMatch(self.fail_condition)


class BaseTestPipelineTask(PipelineTask):
    """A base class for test-utility `PipelineTask` classes that read and write
    mock datasets `runQuantum`.

    Parameters
    ----------
    config : `PipelineTaskConfig`
        The pipeline task config.
    initInputs : `~collections.abc.Mapping`
        The init inputs datasets.
    **kwargs : `~typing.Any`
        Keyword parameters passed to base class constructor.

    Notes
    -----
    This class overrides `runQuantum` to read inputs and write a bit of
    provenance into all of its outputs (always `MockDataset` instances).  It
    can also be configured to raise exceptions on certain data IDs.  It reads
    `MockDataset` inputs and simulates reading inputs of other types by
    creating `MockDataset` inputs from their DatasetRefs.

    Subclasses are responsible for defining connections, but init-input and
    init-output connections are not supported at runtime (they may be present
    as long as the task is never constructed).  All output connections must
    use mock storage classes.  `..Input` and `..PrerequisiteInput` connections
    that do not use mock storage classes will be handled by constructing a
    `MockDataset` from the `~lsst.daf.butler.DatasetRef` rather than actually
    reading them.
    """

    ConfigClass: ClassVar[type[PipelineTaskConfig]] = BaseTestPipelineTaskConfig

    def __init__(
        self,
        *,
        config: BaseTestPipelineTaskConfig,
        initInputs: Mapping[str, Any],
        **kwargs: Any,
    ):
        super().__init__(config=config, **kwargs)
        self.fail_exception: type | None = None
        self.data_id_match = self.config.data_id_match()
        if self.data_id_match:
            self.fail_exception = doImportType(self.config.fail_exception)
            self.memory_required = (
                Quantity(self.config.memory_required) if self.config.memory_required is not None else None
            )
        # Look for, check, and record init-inputs.
        task_connections = self.ConfigClass.ConnectionsClass(config=config)
        mock_dataset_quantum = MockDatasetQuantum(task_label=self.getName(), data_id={}, inputs={})
        for connection_name in task_connections.initInputs:
            input_dataset = initInputs[connection_name]
            if not isinstance(input_dataset, MockDataset):
                raise TypeError(
                    f"Expected MockDataset instance for init-input {self.getName()}.{connection_name}: "
                    f"got {input_dataset!r} of type {type(input_dataset)!r}."
                )
            connection = task_connections.allConnections[connection_name]
            if input_dataset.dataset_type.name != connection.name:
                raise RuntimeError(
                    f"Incorrect dataset type name for init-input {self.getName()}.{connection_name}: "
                    f"got {input_dataset.dataset_type.name!r}, expected {connection.name!r}."
                )
            if input_dataset.storage_class != connection.storageClass:
                raise RuntimeError(
                    f"Incorrect storage class for init-input {self.getName()}.{connection_name}: "
                    f"got {input_dataset.storage_class!r}, expected {connection.storageClass!r}."
                )
            # To avoid very deep provenance we trim inputs to a single
            # level.
            input_dataset.quantum = None
            mock_dataset_quantum.inputs[connection_name] = [input_dataset]
        # Add init-outputs as task instance attributes.
        for connection_name in task_connections.initOutputs:
            connection = task_connections.allConnections[connection_name]
            output_dataset = MockDataset(
                dataset_id=None,  # the task has no way to get this
                dataset_type=SerializedDatasetType(
                    name=connection.name,
                    storageClass=connection.storageClass,
                    dimensions=[],
                ),
                data_id={},
                run=None,  # task also has no way to get this
                quantum=mock_dataset_quantum,
                output_connection_name=connection_name,
                int_value=self.config.int_value,
                str_value=self.config.str_value,
            )
            setattr(self, connection_name, output_dataset)

    config: BaseTestPipelineTaskConfig

    def runQuantum(
        self,
        butlerQC: QuantumContext,
        inputRefs: InputQuantizedConnection,
        outputRefs: OutputQuantizedConnection,
    ) -> None:
        # docstring is inherited from the base class
        quantum = butlerQC.quantum

        _LOG.info("Mocking execution of task '%s' on quantum %s", self.getName(), quantum.dataId)

        if self.config.sleep:
            time.sleep(self.config.sleep)

        assert quantum.dataId is not None, "Quantum DataId cannot be None"

        # Possibly raise an exception.
        if self.data_id_match is not None and self.data_id_match.match(quantum.dataId):
            assert self.fail_exception is not None, "Exception type must be defined"

            if self.memory_required is not None:
                if butlerQC.resources.max_mem < self.memory_required:
                    _LOG.info(
                        "Simulating out-of-memory failure for task '%s' on quantum %s",
                        self.getName(),
                        quantum.dataId,
                    )
                    self._fail(quantum)
            else:
                _LOG.info("Simulating failure of task '%s' on quantum %s", self.getName(), quantum.dataId)
                self._fail(quantum)

        # Populate the bit of provenance we store in all outputs.
        _LOG.info("Reading input data for task '%s' on quantum %s", self.getName(), quantum.dataId)
        mock_dataset_quantum = MockDatasetQuantum(
            task_label=self.getName(), data_id=dict(quantum.dataId.mapping), inputs={}
        )
        for name, refs in inputRefs:
            inputs_list = []
            ref: DatasetRef
            for ref in ensure_iterable(refs):
                if isinstance(ref.datasetType.storageClass, MockStorageClass):
                    input_dataset = butlerQC.get(ref)
                    if isinstance(input_dataset, DeferredDatasetHandle):
                        input_dataset = input_dataset.get()
                    if isinstance(input_dataset, MockDataset):
                        # To avoid very deep provenance we trim inputs to a
                        # single level.
                        input_dataset.quantum = None
                    elif not isinstance(input_dataset, ConvertedUnmockedDataset):
                        raise TypeError(
                            f"Expected MockDataset or ConvertedUnmockedDataset instance for {ref}; "
                            f"got {input_dataset!r} of type {type(input_dataset)!r}."
                        )
                else:
                    input_dataset = MockDataset(
                        dataset_id=ref.id,
                        dataset_type=ref.datasetType.to_simple(),
                        data_id=dict(ref.dataId.mapping),
                        run=ref.run,
                    )
                inputs_list.append(input_dataset)
            mock_dataset_quantum.inputs[name] = inputs_list

        # store mock outputs
        for name, refs in outputRefs:
            for ref in ensure_iterable(refs):
                output = MockDataset(
                    dataset_id=ref.id,
                    dataset_type=ref.datasetType.to_simple(),
                    data_id=dict(ref.dataId.mapping),
                    run=ref.run,
                    quantum=mock_dataset_quantum,
                    output_connection_name=name,
                    int_value=self.config.int_value,
                    str_value=self.config.str_value,
                )
                butlerQC.put(output, ref)

        _LOG.info("Finished mocking task '%s' on quantum %s", self.getName(), quantum.dataId)

    def _fail(self, quantum: Quantum) -> None:
        """Raise the configured exception.

        Parameters
        ----------
        quantum : `lsst.daf.butler.Quantum`
            Quantum producing the error.
        """
        message = f"Simulated failure: task={self.getName()} dataId={quantum.dataId}"
        # Type annotations for optional config fields are broken, so MyPy
        # doesn't think fail_signal could be None.
        if self.config.fail_signal is not None:
            signal.raise_signal(signal.Signals(self.config.fail_signal))
        elif self.fail_exception is AnnotatedPartialOutputsError:  # type: ignore[unreachable]
            # This exception is expected to always chain another.
            try:
                raise MockAlgorithmError(message)
            except AlgorithmError as original:
                error = AnnotatedPartialOutputsError.annotate(original, self, log=self.log)
                raise error from original
        else:
            assert self.fail_exception is not None, "Method should not be called."
            raise self.fail_exception(message)


class MockPipelineDefaultTargetConnections(PipelineTaskConnections, dimensions=()):
    pass


class MockPipelineDefaultTargetConfig(
    PipelineTaskConfig, pipelineConnections=MockPipelineDefaultTargetConnections
):
    pass


class MockPipelineDefaultTargetTask(PipelineTask):
    """A `~lsst.pipe.base.PipelineTask` class used as the default target for
    ``MockPipelineTaskConfig.original``.

    This is effectively a workaround for `lsst.pex.config.ConfigurableField`
    not supporting ``optional=True``, but that is generally a reasonable
    limitation for production code and it wouldn't make sense just to support
    test utilities.
    """

    ConfigClass = MockPipelineDefaultTargetConfig


class MockPipelineTaskConnections(BaseTestPipelineTaskConnections, dimensions=()):
    """A connections class that creates mock connections from the connections
    of a real PipelineTask.

    Parameters
    ----------
    config : `PipelineTaskConfig`
        The config to use for the connection.
    """

    def __init__(self, *, config: MockPipelineTaskConfig):
        self.original: PipelineTaskConnections = config.original.connections.ConnectionsClass(
            config=config.original.value
        )
        self.dimensions.update(self.original.dimensions)
        self.unmocked_dataset_types = frozenset(config.unmocked_dataset_types)
        for name, connection in self.original.allConnections.items():
            if connection.name not in self.unmocked_dataset_types:
                # We register the mock storage class with the global
                # singleton here, but can only put its name in the
                # connection. That means the same global singleton (or one
                # that also has these registrations) has to be available
                # whenever this dataset type is used.
                storage_class_name = MockStorageClass.get_or_register_mock(connection.storageClass).name
                kwargs: dict[str, Any] = {}
                if hasattr(connection, "dimensions"):
                    connection_dimensions = set(connection.dimensions)
                    # Replace the generic "skypix" placeholder with htm7, since
                    # that requires the dataset type to have already been
                    # registered.
                    if "skypix" in connection_dimensions:
                        connection_dimensions.remove("skypix")
                        connection_dimensions.add("htm7")
                    kwargs["dimensions"] = connection_dimensions
                connection = dataclasses.replace(
                    connection,
                    name=get_mock_name(connection.name),
                    storageClass=storage_class_name,
                    **kwargs,
                )
            elif name in self.original.outputs:
                raise ValueError(f"Unmocked dataset type {connection.name!r} cannot be used as an output.")
            elif name in self.original.initInputs:
                raise ValueError(
                    f"Unmocked dataset type {connection.name!r} cannot be used as an init-input."
                )
            elif name in self.original.initOutputs:
                raise ValueError(
                    f"Unmocked dataset type {connection.name!r} cannot be used as an init-output."
                )
            elif (
                connection.name.endswith(METADATA_OUTPUT_CONNECTION_NAME)
                and connection.storageClass == METADATA_OUTPUT_STORAGE_CLASS
            ):
                # Task metadata does not use a mock storage class, because it's
                # written by the system, but it does end up with the _mock_*
                # prefix because the task label does.
                connection = dataclasses.replace(connection, name=get_mock_name(connection.name))
            setattr(self, name, connection)

    def getSpatialBoundsConnections(self) -> Iterable[str]:
        return self.original.getSpatialBoundsConnections()

    def getTemporalBoundsConnections(self) -> Iterable[str]:
        return self.original.getTemporalBoundsConnections()

    def adjustQuantum(
        self,
        inputs: dict[str, tuple[BaseInput, Collection[DatasetRef]]],
        outputs: dict[str, tuple[Output, Collection[DatasetRef]]],
        label: str,
        data_id: DataCoordinate,
    ) -> tuple[
        Mapping[str, tuple[BaseInput, Collection[DatasetRef]]],
        Mapping[str, tuple[Output, Collection[DatasetRef]]],
    ]:
        # Convert the given mappings from the mock dataset types to the
        # original dataset types they were produced from.
        original_inputs = {}
        for connection_name, (_, mock_refs) in inputs.items():
            original_connection = getattr(self.original, connection_name)
            if original_connection.name in self.unmocked_dataset_types:
                refs = mock_refs
            else:
                refs = MockStorageClass.unmock_dataset_refs(mock_refs)
            original_inputs[connection_name] = (original_connection, refs)
        original_outputs = {}
        for connection_name, (_, mock_refs) in outputs.items():
            original_connection = getattr(self.original, connection_name)
            if original_connection.name in self.unmocked_dataset_types:
                refs = mock_refs
            else:
                refs = MockStorageClass.unmock_dataset_refs(mock_refs)
            original_outputs[connection_name] = (original_connection, refs)
        # Call adjustQuantum on the original connections class.
        adjusted_original_inputs, adjusted_original_outputs = self.original.adjustQuantum(
            original_inputs, original_outputs, label, data_id
        )
        # Convert the results back to the mock dataset type.s
        adjusted_inputs = {}
        for connection_name, (original_connection, original_refs) in adjusted_original_inputs.items():
            if original_connection.name in self.unmocked_dataset_types:
                refs = original_refs
            else:
                refs = MockStorageClass.mock_dataset_refs(original_refs)
            adjusted_inputs[connection_name] = (getattr(self, connection_name), refs)
        adjusted_outputs = {}
        for connection_name, (original_connection, original_refs) in adjusted_original_outputs.items():
            if original_connection.name in self.unmocked_dataset_types:
                refs = original_refs
            else:
                refs = MockStorageClass.mock_dataset_refs(original_refs)
            adjusted_outputs[connection_name] = (getattr(self, connection_name), refs)
        return adjusted_inputs, adjusted_outputs


class MockPipelineTaskConfig(BaseTestPipelineTaskConfig, pipelineConnections=MockPipelineTaskConnections):
    """Configuration class for `MockPipelineTask`."""

    original: ConfigurableField = ConfigurableField(
        doc="The original task being mocked by this one.", target=MockPipelineDefaultTargetTask
    )

    unmocked_dataset_types = ListField[str](
        doc=(
            "Names of input dataset types that should be used as-is instead "
            "of being mocked.  May include dataset types not relevant for "
            "this task, which will be ignored."
        ),
        default=(),
        optional=False,
    )


class MockPipelineTask(BaseTestPipelineTask):
    """A test-utility implementation of `PipelineTask` with connections
    generated by mocking those of a real task.

    Notes
    -----
    At present `MockPipelineTask` simply drops any ``initInput`` and
    ``initOutput`` connections present on the original, since `MockDataset`
    creation for those would have to happen in the code that executes the task,
    not in the task itself.  Because `MockPipelineTask` never instantiates the
    mock task (just its connections class), this is a limitation on what the
    mocks can be used to test, not anything deeper.
    """

    ConfigClass: ClassVar[type[PipelineTaskConfig]] = MockPipelineTaskConfig


class DynamicConnectionConfig(Config):
    """A config class that defines a completely dynamic connection."""

    dataset_type_name = Field[str](doc="Name for the dataset type as seen by the butler.", dtype=str)
    dimensions = ListField[str](doc="Dimensions for the dataset type.", dtype=str, default=[])
    storage_class = Field[str](
        doc="Name of the butler storage class for the dataset type.", dtype=str, default="StructuredDataDict"
    )
    is_calibration = Field[bool](doc="Whether this dataset type is a calibration.", dtype=bool, default=False)
    multiple = Field[bool](
        doc="Whether this connection gets or puts multiple datasets for each quantum.",
        dtype=bool,
        default=False,
    )
    mock_storage_class = Field[bool](
        doc="Whether the storage class should actually be a mock of the storage class given.",
        dtype=bool,
        default=True,
    )
    minimum = Field[int](
        doc="Minimum number of datasets per quantum required for this connection.  Ignored for non-inputs.",
        dtype=int,
        default=1,
    )

    def make_connection(self, cls: type[_T]) -> _T:
        storage_class = self.storage_class
        if self.mock_storage_class:
            storage_class = MockStorageClass.get_or_register_mock(storage_class).name
        if issubclass(cls, cT.BaseInput):
            return cls(  # type: ignore
                name=self.dataset_type_name,
                storageClass=storage_class,
                isCalibration=self.is_calibration,
                multiple=self.multiple,
                dimensions=frozenset(self.dimensions),
                minimum=self.minimum,
            )
        elif issubclass(cls, cT.DimensionedConnection):
            return cls(  # type: ignore
                name=self.dataset_type_name,
                storageClass=storage_class,
                isCalibration=self.is_calibration,
                multiple=self.multiple,
                dimensions=frozenset(self.dimensions),
            )
        else:
            return cls(
                name=self.dataset_type_name,
                storageClass=storage_class,
                multiple=self.multiple,
            )


class DynamicTestPipelineTaskConnections(PipelineTaskConnections, dimensions=()):
    """A connections class whose dimensions and connections are wholly
    determined via configuration.

    Parameters
    ----------
    config : `PipelineTaskConfig`
        Config to use for this connections object.
    """

    def __init__(self, *, config: DynamicTestPipelineTaskConfig):
        self.dimensions.update(config.dimensions)
        connection_config: DynamicConnectionConfig
        for connection_name, connection_config in config.init_inputs.items():
            setattr(self, connection_name, connection_config.make_connection(cT.InitInput))
        for connection_name, connection_config in config.init_outputs.items():
            setattr(self, connection_name, connection_config.make_connection(cT.InitOutput))
        for connection_name, connection_config in config.prerequisite_inputs.items():
            setattr(self, connection_name, connection_config.make_connection(cT.PrerequisiteInput))
        for connection_name, connection_config in config.inputs.items():
            setattr(self, connection_name, connection_config.make_connection(cT.Input))
        for connection_name, connection_config in config.outputs.items():
            setattr(self, connection_name, connection_config.make_connection(cT.Output))


class DynamicTestPipelineTaskConfig(
    BaseTestPipelineTaskConfig, pipelineConnections=DynamicTestPipelineTaskConnections
):
    """Configuration for DynamicTestPipelineTask."""

    dimensions = ListField[str](doc="Dimensions for the task's quanta.", dtype=str, default=[])
    init_inputs = ConfigDictField(
        doc=(
            "Init-input connections, keyed by the connection name as seen by the task. "
            "Must be empty if the task will be constructed."
        ),
        keytype=str,
        itemtype=DynamicConnectionConfig,
        default={},
    )
    init_outputs = ConfigDictField(
        doc=(
            "Init-output connections, keyed by the connection name as seen by the task. "
            "Must be empty if the task will be constructed."
        ),
        keytype=str,
        itemtype=DynamicConnectionConfig,
        default={},
    )
    prerequisite_inputs = ConfigDictField(
        doc="Prerequisite input connections, keyed by the connection name as seen by the task.",
        keytype=str,
        itemtype=DynamicConnectionConfig,
        default={},
    )
    inputs = ConfigDictField(
        doc="Regular input connections, keyed by the connection name as seen by the task.",
        keytype=str,
        itemtype=DynamicConnectionConfig,
        default={},
    )
    outputs = ConfigDictField(
        doc="Regular output connections, keyed by the connection name as seen by the task.",
        keytype=str,
        itemtype=DynamicConnectionConfig,
        default={},
    )


class DynamicTestPipelineTask(BaseTestPipelineTask):
    """A test-utility implementation of `PipelineTask` with dimensions and
    connections determined wholly from configuration.
    """

    ConfigClass: ClassVar[type[PipelineTaskConfig]] = DynamicTestPipelineTaskConfig



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
from tempfile import NamedTemporaryFile
from textwrap import dedent

import lsst.pipe.base as pipeBase


class PipelineLoadSubsetTest(unittest.TestCase):
    """Tests for loading subsets of pipelines from YAML files."""

    def setUp(self):
        frozen_pipeline = dedent(
            """
            description: Frozen Pipeline
            tasks:
              isr:
                class: lsst.pipe.base.tests.mocks.DynamicTestPipelineTask
                config:
                  python: |
                    from lsst.pipe.base.tests.mocks import DynamicConnectionConfig
                    config.dimensions = ['exposure', 'detector']
                    config.inputs['input_image'] = DynamicConnectionConfig(
                      dataset_type_name='raw',
                      dimensions={'exposure', 'detector'},
                    )
                    config.outputs['output_image'] = DynamicConnectionConfig(
                      dataset_type_name='detrended',
                      dimensions={'exposure', 'detector'},
                    )
              calibrate:
                class: lsst.pipe.base.tests.mocks.DynamicTestPipelineTask
                config:
                  python: |
                    from lsst.pipe.base.tests.mocks import DynamicConnectionConfig
                    config.dimensions = ['visit', 'detector']
                    config.inputs['input_image'] = DynamicConnectionConfig(
                      dataset_type_name='detrended',
                      dimensions={'exposure', 'detector'},
                    )
                    config.outputs['output_image'] = DynamicConnectionConfig(
                      dataset_type_name='pvi',
                      dimensions={'visit', 'detector'},
                    )
              makeWarp:
                class: lsst.pipe.base.tests.mocks.DynamicTestPipelineTask
                config:
                  python: |
                    from lsst.pipe.base.tests.mocks import DynamicConnectionConfig
                    config.dimensions = ['patch', 'visit']
                    config.inputs['input_image'] = DynamicConnectionConfig(
                      dataset_type_name='pvi',
                      dimensions={'visit', 'detector'},
                    )
                    config.outputs['output_image'] = DynamicConnectionConfig(
                      dataset_type_name='warp',
                      dimensions={'visit', 'patch'},
                    )
              assembleCoadd:
                class: lsst.pipe.base.tests.mocks.DynamicTestPipelineTask
                config:
                  python: |
                    from lsst.pipe.base.tests.mocks import DynamicConnectionConfig
                    config.dimensions = ['patch', 'band']
                    config.inputs['input_image'] = DynamicConnectionConfig(
                      dataset_type_name='warp',
                      dimensions={'visit', 'patch'},
                    )
                    config.outputs['output_image'] = DynamicConnectionConfig(
                      dataset_type_name='coadd',
                      dimensions={'patch', 'band'},
                    )
            """
        )
        self.temp_pipeline_name = ""
        while not self.temp_pipeline_name:
            self.temp_pipeline = NamedTemporaryFile()
            self.temp_pipeline.write(frozen_pipeline.encode())
            self.temp_pipeline.flush()
            if not os.path.exists(self.temp_pipeline.name):
                self.temp_pipeline.close()
            else:
                self.temp_pipeline_name = self.temp_pipeline.name

    def tearDown(self):
        self.temp_pipeline.close()

    def testLoadList(self):
        """Test loading a specific list of labels."""
        labels = ("isr", "makeWarp")
        path = os.path.expandvars(f"{self.temp_pipeline_name}#{','.join(labels)}")
        pipeline = pipeBase.Pipeline.fromFile(path)
        self.assertEqual(set(labels), pipeline._pipelineIR.tasks.keys())

    def testLoadSingle(self):
        """Test loading a specific label."""
        label = "calibrate"
        path = os.path.expandvars(f"{self.temp_pipeline_name}#{label}")
        pipeline = pipeBase.Pipeline.fromFile(path)
        self.assertEqual(set((label,)), pipeline._pipelineIR.tasks.keys())

    def testLoadBoundedRange(self):
        """Test loading a bounded range."""
        path = os.path.expandvars(f"{self.temp_pipeline_name}#calibrate..assembleCoadd")
        pipeline = pipeBase.Pipeline.fromFile(path)
        self.assertEqual(
            {"calibrate", "makeWarp", "assembleCoadd"},
            pipeline._pipelineIR.tasks.keys(),
        )

    def testLoadUpperBound(self):
        """Test loading a range that only has an upper bound."""
        path = os.path.expandvars(f"{self.temp_pipeline_name}#..makeWarp")
        pipeline = pipeBase.Pipeline.fromFile(path)
        self.assertEqual(
            {"isr", "calibrate", "makeWarp"},
            pipeline._pipelineIR.tasks.keys(),
        )

    def testLoadLowerBound(self):
        """Test loading a range that only has a lower bound."""
        path = os.path.expandvars(f"{self.temp_pipeline_name}#makeWarp..")
        pipeline = pipeBase.Pipeline.fromFile(path)
        self.assertEqual(
            {"makeWarp", "assembleCoadd"},
            pipeline._pipelineIR.tasks.keys(),
        )

    def testLabelChecks(self):
        # test a bad list
        path = os.path.expandvars(f"{self.temp_pipeline_name}#FakeLabel")
        with self.assertRaises(ValueError):
            pipeBase.Pipeline.fromFile(path)

        # test a bad end label
        path = os.path.expandvars(f"{self.temp_pipeline_name}#..FakeEndLabel")
        with self.assertRaises(ValueError):
            pipeBase.Pipeline.fromFile(path)

        # test a bad begin label
        path = os.path.expandvars(f"{self.temp_pipeline_name}#FakeBeginLabel..")
        with self.assertRaises(ValueError):
            pipeBase.Pipeline.fromFile(path)

    def testContractRemoval(self):
        path = os.path.expandvars(f"{self.temp_pipeline_name}")
        pipeline = pipeBase.Pipeline.fromFile(path)
        contract = pipeBase.pipelineIR.ContractIR("'visit' in calibrate.dimensions", None)
        pipeline._pipelineIR.contracts.append(contract)
        pipeline = pipeline.subsetFromLabels(pipeBase.LabelSpecifier(labels={"isr"}))
        self.assertEqual(len(pipeline._pipelineIR.contracts), 0)


if __name__ == "__main__":
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Ingester",)

import dataclasses
import logging
import time
import uuid
from collections import defaultdict
from contextlib import AbstractContextManager
from typing import Any, Literal, Self

from lsst.daf.butler import Butler, CollectionType, DatasetRef, DimensionGroup
from lsst.daf.butler.datastore.record_data import DatastoreRecordData
from lsst.daf.butler.registry import ConflictingDefinitionError

from ...pipeline_graph import TaskImportMode
from .._common import DatastoreName
from .._predicted import PredictedQuantumGraphComponents, PredictedQuantumGraphReader
from ._communicators import IngesterCommunicator


@dataclasses.dataclass
class Ingester(AbstractContextManager):
    """A helper class for the provenance aggregator that handles ingestion into
    the central butler repository.
    """

    predicted_path: str
    """Path to the predicted quantum graph."""

    butler_path: str
    """Path or alias to the central butler repository."""

    comms: IngesterCommunicator
    """Communicator object for this worker."""

    predicted: PredictedQuantumGraphComponents = dataclasses.field(init=False)
    """Components of the predicted graph."""

    butler: Butler = dataclasses.field(init=False)
    """Client for the central butler repository."""

    n_datasets_ingested: int = 0
    """Total number of datasets ingested by this invocation."""

    n_datasets_skipped: int = 0
    """Total number of datasets skipped because they were already present."""

    n_producers_pending: int = 0
    """Number of quanta whose outputs are currently pending ingest."""

    refs_pending: defaultdict[DimensionGroup, list[DatasetRef]] = dataclasses.field(
        default_factory=lambda: defaultdict(list)
    )
    """Dataset references pending ingest, grouped by their dimensions."""

    records_pending: dict[DatastoreName, DatastoreRecordData] = dataclasses.field(default_factory=dict)
    """Datastore records pending ingest, grouped by datastore name."""

    already_ingested: set[uuid.UUID] | None = None
    """A set of all dataset IDs already present in the output RUN
    collection.

    If this is not `None`, the ingester is in defensive ingest mode, either
    because it was configured to query for these dataset IDs up front, or
    because a transaction failed due to a dataset already being present.
    """

    last_ingest_time: float = dataclasses.field(default_factory=time.time)
    """POSIX timestamp since the last ingest transaction concluded."""

    def __post_init__(self) -> None:
        self.comms.log.verbose("Reading from predicted quantum graph.")
        with PredictedQuantumGraphReader.open(
            self.predicted_path, import_mode=TaskImportMode.DO_NOT_IMPORT
        ) as reader:
            # We only need the header and pipeline graph.
            self.predicted = reader.components
        if self.comms.config.mock_storage_classes:
            import lsst.pipe.base.tests.mocks  # noqa: F401
        self.comms.log.verbose("Initializing butler.")
        self.butler = Butler.from_config(self.butler_path, writeable=not self.comms.config.dry_run)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Literal[False]:
        try:
            self.butler.close()
        except Exception:
            self.comms.log.exception("An exception occurred during Ingester exit")
        return False

    @property
    def n_datasets_pending(self) -> int:
        """The number of butler datasets currently pending."""
        return sum(len(v) for v in self.refs_pending.values())

    @staticmethod
    def run(predicted_path: str, butler_path: str, comms: IngesterCommunicator) -> None:
        """Run the ingester.

        Parameters
        ----------
        predicted_path : `str`
            Path to the predicted quantum graph.
        butler_path : `str`
            Path or alias to the central butler repository.
        comms : `IngesterCommunicator`
            Communicator for the ingester.

        Notes
        -----
        This method is designed to run as the ``target`` in
        `WorkerFactory.make_worker`.
        """
        with comms, Ingester(predicted_path, butler_path, comms) as ingester:
            ingester.loop()

    def loop(self) -> None:
        """Run the main loop for the ingester."""
        self.comms.log.verbose("Registering collections and dataset types.")
        if not self.comms.config.dry_run:
            if self.comms.config.register_dataset_types:
                self.predicted.pipeline_graph.register_dataset_types(
                    self.butler,
                    include_inputs=False,
                    include_packages=True,
                    include_configs=True,
                    include_logs=True,
                )
            self.butler.collections.register(self.predicted.header.output_run)
            # Updating the output chain cannot happen inside the caching
            # context.
            if self.comms.config.update_output_chain:
                self.update_output_chain()
        with self.butler.registry.caching_context():
            if self.comms.config.defensive_ingest:
                self.fetch_already_ingested()
            self.comms.log.info("Startup completed in %ss.", time.time() - self.last_ingest_time)
            self.last_ingest_time = time.time()
            for ingest_request in self.comms.poll():
                self.n_producers_pending += 1
                self.comms.log.debug(f"Got ingest request for producer {ingest_request.producer_id}.")
                self.update_outputs_pending(refs=ingest_request.refs, records=ingest_request.records)
                if self.n_datasets_pending > self.comms.config.ingest_batch_size:
                    self.ingest()
            self.comms.log.info("All ingest requests received.")
            # We use 'while' in case this fails with a conflict and we switch
            # to defensive mode (should be at most two iterations).
            ingest_start_time = time.time()
            while self.n_datasets_pending:
                n_datasets = self.n_datasets_pending
                self.ingest()
                self.comms.log.verbose(
                    "Gathered %d final datasets in %ss and ingested them in %ss.",
                    n_datasets,
                    ingest_start_time - self.last_ingest_time,
                    time.time() - ingest_start_time,
                )
            if self.n_producers_pending:
                # We can finish with returns pending if we filtered out all of
                # the datasets we started with as already existing.
                self.report()
        self.comms.log_progress(
            logging.INFO,
            f"Ingested {self.n_datasets_ingested} dataset(s); "
            f"skipped {self.n_datasets_skipped} already present.",
        )

    def ingest(self) -> None:
        """Ingest all pending datasets and report success to the supervisor."""
        ingest_start_time = time.time()
        self.comms.log.verbose(
            "Gathered %d datasets from %d quanta in %ss.",
            self.n_datasets_pending,
            self.n_producers_pending,
            ingest_start_time - self.last_ingest_time,
        )
        try:
            if not self.comms.config.dry_run:
                with self.butler.registry.transaction():
                    for refs in self.refs_pending.values():
                        self.butler.registry._importDatasets(refs, expand=False, assume_new=True)
                    self.butler._datastore.import_records(self.records_pending)
            self.last_ingest_time = time.time()
            self.comms.log.verbose(
                "Ingested %d datasets from %d quanta in %ss.",
                self.n_datasets_pending,
                self.n_producers_pending,
                self.last_ingest_time - ingest_start_time,
            )
            self.n_datasets_ingested += self.n_datasets_pending
        except ConflictingDefinitionError:
            if self.already_ingested is None:
                self.comms.log_progress(
                    logging.INFO,
                    "Some outputs seem to have already been ingested; querying for existing datasets and "
                    "switching to defensive ingest mode.",
                )
                self.fetch_already_ingested()
                # We just return instead of trying again immediately because we
                # might have just shrunk the number of pending datasets below
                # the batch threshold.
                return
            else:
                raise
        self.report()
        self.records_pending.clear()
        self.refs_pending.clear()

    def report(self) -> None:
        """Report a successful ingest to the supervisor."""
        self.comms.report_ingest(self.n_producers_pending)
        self.n_producers_pending = 0

    def fetch_already_ingested(self) -> None:
        """Query for the UUIDs of all dataset already present in the output
        RUN collection, and filter and pending datasets accordingly.
        """
        self.comms.log.info("Fetching all UUIDs in output collection %r.", self.predicted.header.output_run)
        self.already_ingested = set(
            self.butler.registry._fetch_run_dataset_ids(self.predicted.header.output_run)
        )
        kept: set[uuid.UUID] = set()
        for dimensions, refs in self.refs_pending.items():
            filtered_refs: list[DatasetRef] = []
            for ref in refs:
                if ref.id not in self.already_ingested:
                    kept.add(ref.id)
                    filtered_refs.append(ref)
                else:
                    self.n_datasets_skipped += 1
            self.refs_pending[dimensions] = filtered_refs
        for datastore_name, datastore_records in list(self.records_pending.items()):
            if (filtered_records := datastore_records.subset(kept)) is not None:
                self.records_pending[datastore_name] = filtered_records
            else:
                del self.records_pending[datastore_name]

    def update_outputs_pending(
        self,
        refs: list[DatasetRef],
        records: dict[DatastoreName, DatastoreRecordData],
    ) -> None:
        """Add an ingest request to the pending-ingest data structures.

        Parameters
        ----------
        refs : `list` [ `lsst.daf.butler.DatasetRef` ]
            Registry information about regular quantum-output datasets.
        records : `dict` [ `str`, \
                `lsst.daf.butler.datastore.record_data.DatastoreRecordData` ]
            Datastore information about the datasets.
        """
        n_given = len(refs)
        if self.already_ingested is not None:
            refs = [ref for ref in refs if ref.id not in self.already_ingested]
            kept = {ref.id for ref in refs}
            self.n_datasets_skipped += n_given - len(kept)
            records = {
                datastore_name: filtered_records
                for datastore_name, original_records in records.items()
                if (filtered_records := original_records.subset(kept)) is not None
            }
        for ref in refs:
            self.refs_pending[ref.datasetType.dimensions].append(ref)
        for datastore_name, datastore_records in records.items():
            if (existing_records := self.records_pending.get(datastore_name)) is not None:
                existing_records.update(datastore_records)
            else:
                self.records_pending[datastore_name] = datastore_records

    def update_output_chain(self) -> None:
        """Update the output CHAINED collection to include the output RUN
        collection (and the inputs, if the output CHAINED collection does not
        exist).

        Notes
        -----
        This method cannot be called inside the registry caching context.
        """
        if self.predicted.header.output is None:
            return
        self.comms.log.info(
            "Updating output collection %s to include %s.",
            self.predicted.header.output,
            self.predicted.header.output_run,
        )
        if self.butler.collections.register(self.predicted.header.output, CollectionType.CHAINED):
            # Chain is new; need to add inputs, but we want to flatten them
            # first.
            if self.predicted.header.inputs:
                flattened = self.butler.collections.query(self.predicted.header.inputs, flatten_chains=True)
                self.butler.collections.extend_chain(self.predicted.header.output, flattened)
        self.butler.collections.prepend_chain(self.predicted.header.output, self.predicted.header.output_run)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""An under-construction version of QuantumGraph and various helper
classes.
"""

from __future__ import annotations

__all__ = (
    "DatasetKey",
    "PrerequisiteDatasetKey",
    "QuantumGraphSkeleton",
    "QuantumKey",
    "TaskInitKey",
)

import dataclasses
from collections import defaultdict
from collections.abc import Iterable, Iterator, MutableMapping, Set
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import networkx

from lsst.daf.butler import (
    Butler,
    DataCoordinate,
    DataIdValue,
    DatasetRef,
    DimensionDataAttacher,
    DimensionDataExtractor,
    DimensionGroup,
    DimensionRecordSet,
)
from lsst.utils.logging import getLogger

if TYPE_CHECKING:
    pass

_LOG = getLogger(__name__)


@dataclasses.dataclass(slots=True, eq=True, frozen=True)
class QuantumKey:
    """Identifier type for quantum keys in a `QuantumGraphSkeleton`."""

    task_label: str
    """Label of the task in the pipeline."""

    data_id_values: tuple[DataIdValue, ...]
    """Data ID values of the quantum.

    Note that keys are fixed given `task_label`, so using only the values here
    speeds up comparisons.
    """

    is_task: ClassVar[Literal[True]] = True
    """Whether this node represents a quantum or task initialization rather
    than a dataset (always `True`).
    """


@dataclasses.dataclass(slots=True, eq=True, frozen=True)
class TaskInitKey:
    """Identifier type for task init keys in a `QuantumGraphSkeleton`."""

    task_label: str
    """Label of the task in the pipeline."""

    is_task: ClassVar[Literal[True]] = True
    """Whether this node represents a quantum or task initialization rather
    than a dataset (always `True`).
    """


@dataclasses.dataclass(slots=True, eq=True, frozen=True)
class DatasetKey:
    """Identifier type for dataset keys in a `QuantumGraphSkeleton`."""

    parent_dataset_type_name: str
    """Name of the dataset type (never a component)."""

    data_id_values: tuple[DataIdValue, ...]
    """Data ID values of the dataset.

    Note that keys are fixed given `parent_dataset_type_name`, so using only
    the values here speeds up comparisons.
    """

    is_task: ClassVar[Literal[False]] = False
    """Whether this node represents a quantum or task initialization rather
    than a dataset (always `False`).
    """

    is_prerequisite: ClassVar[Literal[False]] = False


@dataclasses.dataclass(slots=True, eq=True, frozen=True)
class PrerequisiteDatasetKey:
    """Identifier type for prerequisite dataset keys in a
    `QuantumGraphSkeleton`.

    Unlike regular datasets, prerequisites are not actually required to come
    from a find-first search of `input_collections`, so we don't want to
    assume that the same data ID implies the same dataset.  Happily we also
    don't need to search for them by data ID in the graph, so we can use the
    dataset ID (UUID) instead.
    """

    parent_dataset_type_name: str
    """Name of the dataset type (never a component)."""

    dataset_id_bytes: bytes
    """Dataset ID (UUID) as raw bytes."""

    is_task: ClassVar[Literal[False]] = False
    """Whether this node represents a quantum or task initialization rather
    than a dataset (always `False`).
    """

    is_prerequisite: ClassVar[Literal[True]] = True


type Key = QuantumKey | TaskInitKey | DatasetKey | PrerequisiteDatasetKey


class QuantumGraphSkeleton:
    """An under-construction quantum graph.

    QuantumGraphSkeleton is intended for use inside `QuantumGraphBuilder` and
    its subclasses.

    Parameters
    ----------
    task_labels : `~collections.abc.Iterable` [ `str` ]
        The labels of all tasks whose quanta may be included in the graph, in
        topological order.

    Notes
    -----
    QuantumGraphSkeleton models a bipartite version of the quantum graph, in
    which both quanta and datasets are represented as nodes and each type of
    node only has edges to the other type.

    Square-bracket (`getitem`) indexing returns a mutable mapping of a node's
    flexible attributes.

    The details of the `QuantumGraphSkeleton` API (e.g. which operations
    operate on multiple nodes vs. a single node) are set by what's actually
    needed by current quantum graph generation algorithms.  New variants can be
    added as needed, but adding all operations that *might* be useful for some
    future algorithm seems premature.
    """

    def __init__(self, task_labels: Iterable[str]):
        self._tasks: dict[str, tuple[TaskInitKey, set[QuantumKey]]] = {}
        self._xgraph: networkx.DiGraph = networkx.DiGraph()
        self._global_init_outputs: set[DatasetKey] = set()
        self._dimension_data: dict[str, DimensionRecordSet] = {}
        for task_label in task_labels:
            task_init_key = TaskInitKey(task_label)
            self._tasks[task_label] = (task_init_key, set())
            self._xgraph.add_node(task_init_key)

    def __contains__(self, key: Key) -> bool:
        return key in self._xgraph.nodes

    def __getitem__(self, key: Key) -> MutableMapping[str, Any]:
        return self._xgraph.nodes[key]

    def __iter__(self) -> Iterator[Key]:
        return iter(self._xgraph.nodes)

    @property
    def n_nodes(self) -> int:
        """The total number of nodes of all types."""
        return len(self._xgraph.nodes)

    @property
    def n_edges(self) -> int:
        """The total number of edges."""
        return len(self._xgraph.edges)

    @property
    def has_any_quanta(self) -> bool:
        """Test whether this graph has any quanta."""
        for _ in self.iter_all_quanta():
            return True
        return False

    def has_task(self, task_label: str) -> bool:
        """Test whether the given task is in this skeleton.

        Tasks are only added to the skeleton at initialization, but may be
        removed by `remove_task` if they end up having no quanta.

        Parameters
        ----------
        task_label : `str`
            Task to check for.

        Returns
        -------
        has : `bool`
            `True` if the task is in this skeleton.
        """
        return task_label in self._tasks

    def get_task_init_node(self, task_label: str) -> TaskInitKey:
        """Return the graph node that represents a task's initialization.

        Parameters
        ----------
        task_label : `str`
            The task label to use.

        Returns
        -------
        node : `TaskInitKey`
            The graph node representing this task's initialization.
        """
        return self._tasks[task_label][0]

    def get_quanta(self, task_label: str) -> Set[QuantumKey]:
        """Return the quanta for the given task label.

        Parameters
        ----------
        task_label : `str`
            Label for the task.

        Returns
        -------
        quanta : `~collections.abc.Set` [ `QuantumKey` ]
            A set-like object with the identifiers of all quanta for the given
            task.  *The skeleton object's set of quanta must not be modified
            while iterating over this container; make a copy if mutation during
            iteration is necessary*.
        """
        return self._tasks[task_label][1]

    @property
    def global_init_outputs(self) -> Set[DatasetKey]:
        """The set of dataset nodes that are not associated with any task."""
        return self._global_init_outputs

    def iter_all_quanta(self) -> Iterator[QuantumKey]:
        """Iterate over all quanta from any task, in topological (but otherwise
        unspecified) order.
        """
        for _, quanta in self._tasks.values():
            yield from quanta

    def iter_outputs_of(self, quantum_key: QuantumKey | TaskInitKey) -> Iterator[DatasetKey]:
        """Iterate over the datasets produced by the given quantum.

        Parameters
        ----------
        quantum_key : `QuantumKey` or `TaskInitKey`
            Quantum to iterate over.

        Returns
        -------
        datasets : `~collections.abc.Iterator` of `DatasetKey`
            Datasets produced by the given quanta.
        """
        return self._xgraph.successors(quantum_key)

    def iter_inputs_of(
        self, quantum_key: QuantumKey | TaskInitKey
    ) -> Iterator[DatasetKey | PrerequisiteDatasetKey]:
        """Iterate over the datasets consumed by the given quantum.

        Parameters
        ----------
        quantum_key : `QuantumKey` or `TaskInitKey`
            Quantum to iterate over.

        Returns
        -------
        datasets : `~collections.abc.Iterator` of `DatasetKey` \
                or `PrequisiteDatasetKey`
            Datasets consumed by the given quanta.
        """
        return self._xgraph.predecessors(quantum_key)

    def update(self, other: QuantumGraphSkeleton) -> None:
        """Copy all nodes from ``other`` to ``self``.

        Parameters
        ----------
        other : `QuantumGraphSkeleton`
            Source of nodes. The tasks in ``other`` must be a subset of the
            tasks in ``self`` (this method is expected to be used to populate
            a skeleton for a full from independent-subgraph skeletons).
        """
        for task_label, (_, quanta) in other._tasks.items():
            self._tasks[task_label][1].update(quanta)
        self._xgraph.update(other._xgraph)
        for record_set in other._dimension_data.values():
            self._dimension_data.setdefault(
                record_set.element.name, DimensionRecordSet(record_set.element)
            ).update(record_set)

    def add_quantum_node(self, task_label: str, data_id: DataCoordinate, **attrs: Any) -> QuantumKey:
        """Add a new node representing a quantum.

        Parameters
        ----------
        task_label : `str`
            Name of task.
        data_id : `~lsst.daf.butler.DataCoordinate`
            The data ID of the quantum.
        **attrs : `~typing.Any`
            Additional attributes.
        """
        key = QuantumKey(task_label, data_id.required_values)
        self._xgraph.add_node(key, data_id=data_id, **attrs)
        self._tasks[key.task_label][1].add(key)
        return key

    def add_dataset_node(
        self,
        parent_dataset_type_name: str,
        data_id: DataCoordinate,
        is_global_init_output: bool = False,
        **attrs: Any,
    ) -> DatasetKey:
        """Add a new node representing a dataset.

        Parameters
        ----------
        parent_dataset_type_name : `str`
            Name of the parent dataset type.
        data_id : `~lsst.daf.butler.DataCoordinate`
            The dataset data ID.
        is_global_init_output : `bool`, optional
            Whether this dataset is a global init output.
        **attrs : `~typing.Any`
            Additional attributes for the node.
        """
        key = DatasetKey(parent_dataset_type_name, data_id.required_values)
        self._xgraph.add_node(key, data_id=data_id, **attrs)
        if is_global_init_output:
            assert isinstance(key, DatasetKey), str(key)
            self._global_init_outputs.add(key)
        return key

    def add_prerequisite_node(
        self,
        ref: DatasetRef,
        **attrs: Any,
    ) -> PrerequisiteDatasetKey:
        """Add a new node representing a prerequisite input dataset.

        Parameters
        ----------
        ref : `~lsst.daf.butler.DatasetRef`
            The dataset ref of the prerequisite.
        **attrs : `~typing.Any`
            Additional attributes for the node.
        """
        key = PrerequisiteDatasetKey(ref.datasetType.name, ref.id.bytes)
        self._xgraph.add_node(key, data_id=ref.dataId, ref=ref, **attrs)
        return key

    def remove_quantum_node(self, key: QuantumKey, remove_outputs: bool) -> None:
        """Remove a node representing a quantum.

        Parameters
        ----------
        key : `QuantumKey`
            Identifier for the node.
        remove_outputs : `bool`
            If `True`, also remove all dataset nodes produced by this quantum.
            If `False`, any such dataset nodes will become overall inputs.
        """
        _, quanta = self._tasks[key.task_label]
        quanta.remove(key)
        if remove_outputs:
            to_remove = list(self._xgraph.successors(key))
            to_remove.append(key)
            self._xgraph.remove_nodes_from(to_remove)
        else:
            self._xgraph.remove_node(key)

    def remove_dataset_nodes(self, keys: Iterable[DatasetKey | PrerequisiteDatasetKey]) -> None:
        """Remove nodes representing datasets.

        Parameters
        ----------
        keys : `~collections.abc.Iterable` of `DatasetKey`\
                or `PrerequisiteDatasetKey`
            Nodes to remove.
        """
        self._xgraph.remove_nodes_from(keys)

    def remove_task(self, task_label: str) -> None:
        """Fully remove a task from the skeleton.

        All init-output datasets and quanta for the task must already have been
        removed.

        Parameters
        ----------
        task_label : `str`
            Name of task to remove.
        """
        task_init_key, quanta = self._tasks.pop(task_label)
        assert not quanta, "Cannot remove task unless all quanta have already been removed."
        assert not list(self._xgraph.successors(task_init_key))
        self._xgraph.remove_node(task_init_key)

    def add_input_edges(
        self,
        task_key: QuantumKey | TaskInitKey,
        dataset_keys: Iterable[DatasetKey | PrerequisiteDatasetKey],
    ) -> None:
        """Add edges connecting datasets to a quantum that consumes them.

        Parameters
        ----------
        task_key : `QuantumKey` or `TaskInitKey`
            Quantum to connect.
        dataset_keys : `~collections.abc.Iterable` of `DatasetKey`\
                or `PrequisiteDatasetKey`
            Datasets to join to the quantum.

        Notes
        -----
        This must only be called if the task node has already been added.
        Use `add_input_edge` if this cannot be assumed.

        Dataset nodes that are not already present will be created.
        """
        assert task_key in self._xgraph, str(task_key)
        self._xgraph.add_edges_from((dataset_key, task_key) for dataset_key in dataset_keys)

    def remove_input_edges(
        self,
        task_key: QuantumKey | TaskInitKey,
        dataset_keys: Iterable[DatasetKey | PrerequisiteDatasetKey],
    ) -> None:
        """Remove edges connecting datasets to a quantum that consumes them.

        Parameters
        ----------
        task_key : `QuantumKey` or `TaskInitKey`
            Quantum to disconnect.
        dataset_keys : `~collections.abc.Iterable` of `DatasetKey`\
                or `PrequisiteDatasetKey`
            Datasets to remove from the quantum.
        """
        self._xgraph.remove_edges_from((dataset_key, task_key) for dataset_key in dataset_keys)

    def add_input_edge(
        self,
        task_key: QuantumKey | TaskInitKey,
        dataset_key: DatasetKey | PrerequisiteDatasetKey,
        ignore_unrecognized_quanta: bool = False,
    ) -> bool:
        """Add an edge connecting a dataset to a quantum that consumes it.

        Parameters
        ----------
        task_key : `QuantumKey` or `TaskInitKey`
            Identifier for the quantum node.
        dataset_key : `DatasetKey` or `PrerequisiteKey`
            Identifier for the dataset node.
        ignore_unrecognized_quanta : `bool`, optional
            If `False`, do nothing if the quantum node is not already present.
            If `True`, the quantum node is assumed to be present.

        Returns
        -------
        added : `bool`
            `True` if an edge was actually added, `False` if the quantum was
            not recognized and the edge was not added as a result.

        Notes
        -----
        Dataset nodes that are not already present will be created.
        """
        if ignore_unrecognized_quanta and task_key not in self._xgraph:
            return False
        self._xgraph.add_edge(dataset_key, task_key)
        return True

    def add_output_edge(self, task_key: QuantumKey | TaskInitKey, dataset_key: DatasetKey) -> None:
        """Add an edge connecting a dataset to the quantum that produces it.

        Parameters
        ----------
        task_key : `QuantumKey` or `TaskInitKey`
            Identifier for the quantum node.  Must identify a node already
            present in the graph.
        dataset_key : `DatasetKey`
            Identifier for the dataset node.  Must identify a node already
            present in the graph.
        """
        assert task_key in self._xgraph, str(task_key)
        assert dataset_key in self._xgraph, str(dataset_key)
        self._xgraph.add_edge(task_key, dataset_key)

    def remove_output_edge(self, dataset_key: DatasetKey) -> None:
        """Remove the edge connecting a dataset to the quantum that produces
        it.

        Parameters
        ----------
        dataset_key : `DatasetKey`
            Identifier for the dataset node.  Must identify a node already
            present in the graph.
        """
        (task_key,) = self._xgraph.predecessors(dataset_key)
        assert dataset_key in self._xgraph, str(dataset_key)
        self._xgraph.remove_edge(task_key, dataset_key)

    def remove_orphan_datasets(self) -> None:
        """Remove any dataset nodes that do not have any edges."""
        for orphan in list(networkx.isolates(self._xgraph)):
            if not orphan.is_task and orphan not in self._global_init_outputs:
                self._xgraph.remove_node(orphan)

    def extract_overall_inputs(self) -> dict[DatasetKey | PrerequisiteDatasetKey, DatasetRef]:
        """Find overall input datasets.

        Returns
        -------
        datasets : `dict` [ `DatasetKey` or `PrerequisiteDatasetKey`, \
                `~lsst.daf.butler.DatasetRef` ]
            Overall-input datasets, including prerequisites and init-inputs.
        """
        result = {}
        for generation in networkx.algorithms.topological_generations(self._xgraph):
            for dataset_key in generation:
                if dataset_key.is_task:
                    continue
                if (ref := self.get_dataset_ref(dataset_key)) is None:
                    raise AssertionError(
                        f"Logic bug in QG generation: dataset {dataset_key} was never resolved."
                    )
                result[dataset_key] = ref
            break
        return result

    def set_dataset_ref(
        self, ref: DatasetRef, key: DatasetKey | PrerequisiteDatasetKey | None = None
    ) -> None:
        """Associate a dataset node with a `~lsst.daf.butler.DatasetRef`
        instance.

        Parameters
        ----------
        ref : `~lsst.daf.butler.DatasetRef`
            `~lsst.daf.butler.DatasetRef` to associate with the node.
        key : `DatasetKey` or `PrerequisiteDatasetKey`, optional
            Identifier for the graph node.  If not provided, a `DatasetKey`
            is constructed from the dataset type name and data ID of ``ref``.
        """
        if key is None:
            key = DatasetKey(ref.datasetType.name, ref.dataId.required_values)
        self._xgraph.nodes[key]["ref"] = ref

    def set_output_for_skip(self, ref: DatasetRef) -> None:
        """Associate a dataset node with a `~lsst.daf.butler.DatasetRef` that
        represents an existing output in a collection where such outputs can
        cause a quantum to be skipped.

        Parameters
        ----------
        ref : `~lsst.daf.butler.DatasetRef`
            `~lsst.daf.butler.DatasetRef` to associate with the node.
        """
        key = DatasetKey(ref.datasetType.name, ref.dataId.required_values)
        self._xgraph.nodes[key]["output_for_skip"] = ref

    def set_output_in_the_way(self, ref: DatasetRef) -> None:
        """Associate a dataset node with a `~lsst.daf.butler.DatasetRef` that
        represents an existing output in the output RUN collection.

        Parameters
        ----------
        ref : `~lsst.daf.butler.DatasetRef`
            `~lsst.daf.butler.DatasetRef` to associate with the node.
        """
        key = DatasetKey(ref.datasetType.name, ref.dataId.required_values)
        self._xgraph.nodes[key]["output_in_the_way"] = ref

    def get_dataset_ref(self, key: DatasetKey | PrerequisiteDatasetKey) -> DatasetRef | None:
        """Return the `~lsst.daf.butler.DatasetRef` associated with the given
        node.

        This does not return "output for skip" and "output in the way"
        datasets.

        Parameters
        ----------
        key : `DatasetKey` or `PrerequisiteDatasetKey`
            Identifier for the graph node.

        Returns
        -------
        ref : `~lsst.daf.butler.DatasetRef` or `None`
            Dataset reference associated with the node.
        """
        return self._xgraph.nodes[key].get("ref")

    def get_output_for_skip(self, key: DatasetKey) -> DatasetRef | None:
        """Return the `~lsst.daf.butler.DatasetRef` associated with the given
        node in a collection where it could lead to a quantum being skipped.

        Parameters
        ----------
        key : `DatasetKey`
            Identifier for the graph node.

        Returns
        -------
        ref : `~lsst.daf.butler.DatasetRef` or `None`
            Dataset reference associated with the node.
        """
        return self._xgraph.nodes[key].get("output_for_skip")

    def get_output_in_the_way(self, key: DatasetKey) -> DatasetRef | None:
        """Return the `~lsst.daf.butler.DatasetRef` associated with the given
        node in the output RUN collection.

        Parameters
        ----------
        key : `DatasetKey`
            Identifier for the graph node.

        Returns
        -------
        ref : `~lsst.daf.butler.DatasetRef` or `None`
            Dataset reference associated with the node.
        """
        return self._xgraph.nodes[key].get("output_in_the_way")

    def discard_output_in_the_way(self, key: DatasetKey) -> None:
        """Drop any `~lsst.daf.butler.DatasetRef` associated with this node in
        the output RUN collection.

        Does nothing if there is no such `~lsst.daf.butler.DatasetRef`.

        Parameters
        ----------
        key : `DatasetKey`
            Identifier for the graph node.
        """
        self._xgraph.nodes[key].pop("output_in_the_way", None)

    def set_data_id(self, key: Key, data_id: DataCoordinate) -> None:
        """Set the data ID associated with a node.

        This updates the data ID in any `~lsst.daf.butler.DatasetRef` objects
        associated with the node via `set_ref`, `set_output_for_skip`, or
        `set_output_in_the_way` as well, assuming it is an expanded version
        of the original data ID.

        Parameters
        ----------
        key : `Key`
            Identifier for the graph node.
        data_id : `~lsst.daf.butler.DataCoordinate`
            Data ID for the node.
        """
        state: MutableMapping[str, Any] = self._xgraph.nodes[key]
        state["data_id"] = data_id
        ref: DatasetRef | None
        if (ref := state.get("ref")) is not None:
            state["ref"] = ref.expanded(data_id)
        output_for_skip: DatasetRef | None
        if (output_for_skip := state.get("output_for_skip")) is not None:
            state["output_for_skip"] = output_for_skip.expanded(data_id)
        output_in_the_way: DatasetRef | None
        if (output_in_the_way := state.get("output_in_the_way")) is not None:
            state["output_in_the_way"] = output_in_the_way.expanded(data_id)

    def get_data_id(self, key: Key) -> DataCoordinate:
        """Return the full data ID for a quantum or dataset, if available.

        Parameters
        ----------
        key : `Key`
            Identifier for the graph node.

        Returns
        -------
        data_id : `~lsst.daf.butler.DataCoordinate`
            Expanded data ID for the node, if one is available.

        Raises
        ------
        KeyError
            Raised if this node does not have an expanded data ID.
        """
        return self._xgraph.nodes[key]["data_id"]

    def attach_dimension_records(
        self,
        butler: Butler,
        dimensions: DimensionGroup,
        dimension_records: Iterable[DimensionRecordSet] = (),
    ) -> None:
        """Attach dimension records to the data IDs in the skeleton.

        This both attaches records to data IDs in the skeleton and aggregates
        any existing records on data IDS, so `get_dimension_data` returns all
        dimension records used in the skeleton.  It can be called multiple
        times.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            Butler to use to query for missing dimension records.
        dimensions : `lsst.daf.butler.DimensionGroup`
            Superset of all of the dimensions of all data IDs.
        dimension_records : `~collections.abc.Iterable` [ \
                `lsst.daf.butler.DimensionRecordSet` ], optional
            Iterable of sets of dimension records to attach.
        """
        for record_set in dimension_records:
            self._dimension_data.setdefault(
                record_set.element.name, DimensionRecordSet(record_set.element)
            ).update(record_set)
        # Group all nodes by data ID (and dimensions of data ID).
        data_ids_to_expand: defaultdict[DimensionGroup, defaultdict[DataCoordinate, list[Key]]] = defaultdict(
            lambda: defaultdict(list)
        )
        extractor = DimensionDataExtractor.from_dimension_group(dimensions)
        data_id: DataCoordinate | None
        for node_key in self:
            if data_id := self[node_key].get("data_id"):
                if data_id.hasRecords():
                    extractor.update([data_id])
                else:
                    data_ids_to_expand[data_id.dimensions][data_id].append(node_key)
        # Add records we extracted from data IDs that were already expanded, in
        # case other nodes want them.
        for record_set in extractor.records.values():
            self._dimension_data.setdefault(
                record_set.element.name, DimensionRecordSet(record_set.element)
            ).update(record_set)
        attacher = DimensionDataAttacher(records=self._dimension_data.values(), dimensions=dimensions)
        for dimensions, data_ids in data_ids_to_expand.items():
            with butler.query() as query:
                # Butler query will be used as-needed to get dimension records
                # (from prerequisites) we didn't fetch in advance.  These are
                # cached in the attacher so we don't look them up multiple
                # times.
                expanded_data_ids = attacher.attach(dimensions, data_ids.keys(), query=query)
            for expanded_data_id, node_keys in zip(expanded_data_ids, data_ids.values()):
                for node_key in node_keys:
                    self.set_data_id(node_key, expanded_data_id)
        # Hold on to any records that we had to query for or extracted.
        self._dimension_data = attacher.records

    def get_dimension_data(self) -> list[DimensionRecordSet]:
        """Return the dimension records attached to data IDs."""
        return list(self._dimension_data.values())
